package com.monkeycoder.monkeynes.gui;

import com.monkeycoder.monkeynes.hardware.*;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;

/**
 * This is a GUI dialog that allows the user to configure the
 * active codes of the Game Genie system.
 * <br><br> 
 * Originally part of the MonkeyNES project.
 *
 * @author David Stephens
 */
public class GameGenieGui
{
  private Frame                parentFrame                      = null;
  private JDialog              mainDialog                       = null;

  private JTable               codesTable                       = null;
  private JTableHeader         codesTableHeader                 = null;
  private JScrollPane          tableScroll                      = null;
  private GameGenieTableModel  ggtm                             = null;
  
  private JPanel               bottomPanel                      = new JPanel( new BorderLayout() );
  private JPanel               inputPanel                       = new JPanel( new GridLayout(5,3) );

  private JTextField           codeText                         = new JTextField( "", 8 );
  private JTextField           addressText                      = new JTextField( "", 4 );
  private JTextField           valueText                        = new JTextField( "", 2 );
  private JTextField           testText                         = new JTextField( "", 2 );
 
  private JButton              addButton                        = new JButton( "Add" );
  private JButton              removeButton                     = new JButton( "Remove" );
  
  /** 
   * Creates and configures a new GameGenieGui.
   * Note that the window is constructed but not displayed.
   * @param inParentFrame The parent of this dialog.
   */    
  public GameGenieGui( Frame inParentFrame )
  {
    parentFrame = inParentFrame;
    
    // Dialog
    mainDialog = new JDialog( parentFrame, "Options: Game Genie", false );
    
    // Put together the active codes table
    String[] columnNames = { "Code", "Address", "Value", "Test" };
    ggtm             = new GameGenieTableModel( columnNames );
    codesTable       = new JTable( ggtm );
    codesTableHeader = codesTable.getTableHeader();
    tableScroll      = new JScrollPane( codesTable );    
    codesTable.setPreferredScrollableViewportSize(new Dimension(325, 200));

    // Table Test Data    
    ArrayList a = new ArrayList();
    GameGenieCode g = GameGenieCode.translate( "GOSSIP" );
    a.add( g );
    g = GameGenieCode.translate( "ZEXPYGLA" );
    a.add( g );
    ggtm.setDataSet( a );
    
    // Build the input panel
    inputPanel.add( new JLabel( "Code" ) );
    inputPanel.add( new JLabel() );
    inputPanel.add( codeText );
    inputPanel.add( new JLabel() );
    inputPanel.add( GuiHelper.flowWrapper(FlowLayout.CENTER, new JLabel( "-OR-" ) ) );
    inputPanel.add( new JLabel() );
    inputPanel.add( new JLabel( "Address" ) );
    inputPanel.add( new JLabel() );
    inputPanel.add( addressText );
    inputPanel.add( new JLabel( "Value" ) );
    inputPanel.add( new JLabel() );
    inputPanel.add( valueText );
    inputPanel.add( new JLabel( "Test" ) );
    inputPanel.add( new JLabel() );
    inputPanel.add( testText );
    
    // Build the bottom panel
    bottomPanel.add( GuiHelper.flowWrapper(FlowLayout.CENTER, removeButton ), BorderLayout.NORTH );
    bottomPanel.add( inputPanel,                                              BorderLayout.CENTER );
    bottomPanel.add( GuiHelper.flowWrapper(FlowLayout.CENTER, addButton ),    BorderLayout.SOUTH );
    
    // Build the main view
    mainDialog.getContentPane().add( tableScroll,     BorderLayout.CENTER  );
    mainDialog.getContentPane().add( bottomPanel,     BorderLayout.SOUTH   );

    // Main window closing listener
    mainDialog.addWindowListener( new WindowAdapter()
    {
      public void windowClosing( WindowEvent we )
      {
        mainDialog.hide();
      }
    } );

    codesTable.addMouseListener( new MouseAdapter()
    {
      public void mouseClicked( MouseEvent me )
      {
        // Transfer the values from the clicked row into the address, value, and test fields
        int row = codesTable.getSelectedRow();
        
        if( row != -1 )
        {
          codeText.setText( "" );
          addressText.setText( (String)ggtm.getValueAt( row, 1 ) );
          valueText.setText( (String)ggtm.getValueAt( row, 2 ) );
          
          if( ((String)ggtm.getValueAt( row, 3 )).equals( "N/A" ) )
            testText.setText( "" );
          else
            testText.setText( (String)ggtm.getValueAt( row, 3 ) );
        }    
      }
    } );
    
    // Add button listener
    addButton.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        int           address = 0;
        int           value   = 0;
        int           test    = 0;
        GameGenieCode code    = null;
        
        if( !codeText.getText().equals( "" ) )
        {
          code = GameGenieCode.translate( codeText.getText() );  
        }
        else
        {
          if( testText.getText().equals( "" ) )
          {
            try
            {
              address = Integer.parseInt( addressText.getText(), 16 );
              value   = Integer.parseInt( valueText.getText(),   16 );
            }
            catch( NumberFormatException nfe )
            {
              return;
            }
            
            code = new GameGenieCode( address, value );
          }
          else
          {
            try
            {
              address = Integer.parseInt( addressText.getText(), 16 );
              value   = Integer.parseInt( valueText.getText(), 16 );
              test    = Integer.parseInt( testText.getText(), 16 );
            }
            catch( NumberFormatException nfe )
            {
              return;
            }
            
            code = new GameGenieCode( address, value, test );
          }  
        }
        
        if( code != null )
          ggtm.addRow( code );
      }
    } );
    
    // Remove button listener
    removeButton.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        int row = codesTable.getSelectedRow();
        if( row != -1 )
        {
          ggtm.removeRow( row );
          
          // Clear the fields since clicking a row loads them
          codeText.setText( "" );
          addressText.setText( "" );
          valueText.setText( "" );
          testText.setText( "" );
        }
      }
    } );
    
    // codeText field key listener
    codeText.addKeyListener( new KeyAdapter()
    {
      public void keyPressed( KeyEvent ke )
      {
        addressText.setText( "" );
        valueText.setText( "" );
        testText.setText( "" );
      }
    } );
    
    // addressText field key listener
    addressText.addKeyListener( new KeyAdapter()
    {
      public void keyPressed( KeyEvent ke )
      {
        codeText.setText( "" );
      }
    } );
    
    // valueText field key listener
    valueText.addKeyListener( new KeyAdapter()
    {
      public void keyPressed( KeyEvent ke )
      {
        codeText.setText( "" );
      }
    } );  
    
    // testText field key listener
    testText.addKeyListener( new KeyAdapter()
    {
      public void keyPressed( KeyEvent ke )
      {
        codeText.setText( "" );
      }
    } );  
  } // End of constructor ----------------------------------------------------------

  /**
   * Displays the GameGenieGui.
   */  
  public void show()
  {
    // Position the window
    Dimension parentSize = parentFrame.getSize();
    Point     parentLoc  = parentFrame.getLocation();
    Dimension mySize     = new Dimension( 350, 450 );
    Point     myLoc      = new Point( (parentLoc.x + (parentSize.width / 2) - (mySize.width / 2 )), 
                                      (parentLoc.y + (parentSize.height / 2) - (mySize.height / 2 )) );

    // Fix the location in case the positioning put part of the window off the screen                                      
    if( myLoc.x < 0 ) myLoc.x = 0;                                      
    if( myLoc.y < 0 ) myLoc.y = 0;
    
    // Size and position the window
    mainDialog.setLocation( myLoc );
    mainDialog.setSize( mySize );
    
    // Show the whole mess...
    mainDialog.show();
  }

}