package com.monkeycoder.monkeynes.gui;

import com.monkeycoder.monkeynes.hardware.*;

import java.util.*;
import javax.swing.table.*;
import javax.swing.event.*;

/**
 * This is the table model used for the display of GameGenieCodes by
 * the GameGenieGui.
 * <br><br> 
 * Originally part of the MonkeyNES project.
 *
 * @author David Stephens
 */
public class GameGenieTableModel extends AbstractTableModel 
{
  private String[]   columnNames = null;
  private Object[][] data        = null;

  /** 
   * Creates a new GameGenieTableModel with no headers or data.
   */
  public GameGenieTableModel()
  {
    columnNames = new String[0];
    data        = new Object[0][0];
  }

  /** 
   * Creates a new GameGenieTableModel with the supplied headers.
   * @param inColNames A string array containing column headers labels.
   */
  public GameGenieTableModel( String[] inColNames )
  {
    columnNames = inColNames;
    data        = new Object[0][0];
  }

  /** 
   * Creates a new GameGenieTableModel with the supplied headers and raw data.
   * @param inColNames A string array containing column headers labels.
   * @param inData Object data for each cell of the table.
   */
  public GameGenieTableModel( String[] inColNames, Object[][] inData )
  {
    columnNames = inColNames;
    data        = inData;
  }

  /** 
   * Creates a new GameGenieTableModel with the supplied headers and data.
   * @param inColNames A string array containing column headers labels.
   * @param inData An ArrayList expected to contain GameGenieCode objects.
   */
  public GameGenieTableModel( String[] inColNames, ArrayList inData )
  {
    columnNames = inColNames;
    setDataSet( inData );
  }
  
  /**
   * Builds an ArrayList full of all the GameGenieCode objects currently in the table.
   * @return An ArrayList containing GameGenieCode objects.
   */
  public ArrayList getDataSet()
  {
    // Reconstruct the data from the table into an ArrayList
    ArrayList tempList = new ArrayList();
    
    for( int i = 0; i < data.length; i++ )
    {
      GameGenieCode tempRec = GameGenieCode.translate( (String)data[i][0] );
     
      tempList.add( tempRec );
    }
    
    return tempList;
  }
  
  /**
   * Replaces the data used with new data.
   * @param inData An ArrayList expected to contain GameGenieCode objects.
   */
  public void setDataSet( ArrayList inData )
  {
    // Load the data array from the ArrayList
    data = new Object[inData.size()][4];
    for( int i = 0; i < inData.size(); i++ )
    {
      GameGenieCode tempRecord = (GameGenieCode)inData.get( i );
      data[i][0] = tempRecord.getCode();
      data[i][1] = Integer.toString( tempRecord.getAddress(), 16 ).toUpperCase();
      data[i][2] = Integer.toString( tempRecord.getValue(), 16 ).toUpperCase();
      
      if( tempRecord.usesTest() )
        data[i][3] = Integer.toString( tempRecord.getTest(), 16 ).toUpperCase();
      else
        data[i][3] = "N/A";
    }    
    
    fireTableDataChanged();
  }
    
  /**
   * Returns the number of columns in this table.  Should always be 4.
   * @return The number of columns in this table. 
   */    
  public int getColumnCount() 
  {
    return columnNames.length;
  }

  /**
   * Returns the number of data rows in this table.
   * @return The number of data rows in this table. 
   */
  public int getRowCount() 
  {
    return data.length;
  }

  /**
   * Returns the column header name for the requested column.
   * @param col The requested column.
   * @return The column header name for the requested column.
   */
  public String getColumnName( int col ) 
  {
    return columnNames[col];
  }

  /**
   * Returns the data at the specified table coordinates.
   * @param row The row coordinate.
   * @param col The column coordinate.
   * @return The data at the specified table coordinates.
   */
  public Object getValueAt( int row, int col ) 
  {
    return data[row][col];
  }

  /**
   * Returns the object class of the requested column.
   * @param col The requested column.
   * @return The object class of the requested column.
   */
  public Class getColumnClass( int col ) 
  {
    return getValueAt( 0, col ).getClass();
  }

  /**
   * Returns whether or not the data at the specified table coordinates is editable.
   * @param row The row coordinate.
   * @param col The column coordinate.
   * @return Whether or not the data at the specified table coordinates is editable.
   */
  public boolean isCellEditable( int row, int col ) 
  {
    return false;
  }

  /**
   * Sets the raw data at the specified table coordinates.
   * @param value The raw data.
   * @param row The row coordinate.
   * @param col The column coordinate.
   */
  public void setValueAt( Object value, int row, int col ) 
  {
    data[row][col] = value;
    fireTableCellUpdated(row, col);
  }

  /**
   * Adds a row to the bottom of the table.
   * @param inCode The GameGenieCode that will be used to add a row to the table.
   */  
  public void addRow( GameGenieCode inCode )
  {
    ArrayList tempData = getDataSet();
    tempData.add( inCode );
    setDataSet( tempData );
    
    fireTableDataChanged();
  }

  /**
   * Removes a row from the table.
   * @param row The index of the row to remove.
   */    
  public void removeRow( int row )
  {
    ArrayList tempData = getDataSet();
    tempData.remove( row );
    setDataSet( tempData );
    
    fireTableDataChanged();
  }
}
