package com.monkeycoder.monkeynes.gui;

import com.monkeycoder.monkeynes.base.*;
import com.monkeycoder.monkeynes.hardware.*;
import com.monkeycoder.monkeynes.hardware.mapper.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;

/**
 * This is the main window of MonkeyNES.
 * <br><br> 
 * Originally part of the MonkeyNES project.
 *
 * @author David Stephens
 */
public class MonkeyNesGui
{
  private JFrame             mainFrame                        = new JFrame();
  private JFileChooser       fileChooser                      = null;
  private GenericFileFilter  romTypeFilter                    = new GenericFileFilter( "nes", "NES ROM Files (*.nes)", true );
  
  // Other windows
  private OptionsCpuGui      optionsCpuWindow                 = new OptionsCpuGui( mainFrame );
  private OptionsVideoGui    optionsVideoWindow               = new OptionsVideoGui( mainFrame );
  private OptionsSoundGui    optionsSoundWindow               = new OptionsSoundGui( mainFrame );
  private OptionsInputGui    optionsInputWindow               = new OptionsInputGui( mainFrame );
  private GameGenieGui       gameGenieWindow                  = new GameGenieGui( mainFrame );
  private HelpAboutGui       helpAboutWindow                  = new HelpAboutGui( mainFrame );                  
  // See Tool Windows below
  
  // Menu System
  private JMenuBar           mainMenu                         = new JMenuBar();
  private JMenu              fileMenu                         = new JMenu( "File" );
  private JMenuItem          fileOpenMenuItem                 = new JMenuItem( "Open ROM", KeyEvent.VK_O );
  private JMenuItem          fileCloseMenuItem                = new JMenuItem( "Close ROM", KeyEvent.VK_C );
  private JMenuItem          fileExitMenuItem                 = new JMenuItem( "Exit", KeyEvent.VK_X );
  private JMenu              cpuMenu                          = new JMenu( "CPU" );
  private JCheckBoxMenuItem  cpuPowerMenuItem                 = new JCheckBoxMenuItem( "Power", false );
  private JMenuItem          cpuResetMenuItem                 = new JMenuItem( "Reset", KeyEvent.VK_R );
  private JMenu              optMenu                          = new JMenu( "Options" );
  private JMenuItem          optCpuMenuItem                   = new JMenuItem( "CPU...", KeyEvent.VK_C );
  private JMenuItem          optVideoMenuItem                 = new JMenuItem( "Video...", KeyEvent.VK_V );
  private JMenuItem          optSoundMenuItem                 = new JMenuItem( "Sound...", KeyEvent.VK_S );
  private JMenuItem          optInputMenuItem                 = new JMenuItem( "Input...", KeyEvent.VK_I );
  private JMenuItem          optGameGenieMenuItem             = new JMenuItem( "Game Genie...", KeyEvent.VK_G );
  private JMenu              toolMenu                         = new JMenu( "Tools" );
  private JMenuItem          toolCpuStateViewerMenuItem       = new JMenuItem( "CPU State Viewer" );
  private JMenuItem          toolOpCodeTestShellMenuItem      = new JMenuItem( "OpCode Test Shell" );
  private JMenuItem          toolPatternViewerMenuItem        = new JMenuItem( "Pattern Table Viewer" );
  private JMenuItem          toolPrgRomViewerMenuItem         = new JMenuItem( "PRG-ROM Disassembly Viewer" );
  private JMenuItem          toolPrgRomHexViewerMenuItem      = new JMenuItem( "PRG-ROM Hex Viewer" );
  private JMenuItem          toolLoadSmbMenuItem              = new JMenuItem( "Quick Load SMB" );
  private JMenu              helpMenu                         = new JMenu( "Help" );
  private JMenuItem          helpAboutMenuItem                = new JMenuItem( "About", KeyEvent.VK_A );

  // Status Bar
  private JPanel             statusPanel                      = new JPanel( new BorderLayout() );
  private JPanel             lightPanel                       = new JPanel( new FlowLayout( FlowLayout.RIGHT, 1, 0 ) );
  private JLabel             statusLabel                      = new JLabel( "Idle" );
  private StateLight         romLoadedLight                   = new StateLight( new Color(0,255,0), new Color(0,128,0), new Color(0,128,0), new Color(0,255,0) );
  private StateLight         powerOnLight                     = new StateLight( new Color(255,0,0), new Color(128,0,0), new Color(128,0,0), new Color(255,0,0) );

  // Hardware
  private VideoScreen        outputScreen                     = new VideoScreen();
  private NesHardware        nes                              = new NesHardware();
  private NesJoypad          pad1                             = new NesJoypad();
  private NesJoypad          pad2                             = new NesJoypad();
  private NesRom             rom                              = null;
  
  // UI
  private NesControllerKeyListener padListener                = new NesControllerKeyListener( pad1, pad2 );
  
  // Tool Windows
  private CpuStateViewerGui     cpuStateViewerWindow          = new CpuStateViewerGui( mainFrame, nes.getCpu() );
  private OpCodeTestShellGui    opCodeTestShellWindow         = new OpCodeTestShellGui( mainFrame, nes );
  private PatternTableViewerGui patternTableViewerWindow      = new PatternTableViewerGui( mainFrame, nes.getVideoMemory() );
  private PrgRomViewerGui       prgRomViewerWindow            = new PrgRomViewerGui( mainFrame, nes.getMainMemory() );
  private PrgRomHexViewerGui    prgRomHexViewerWindow         = new PrgRomHexViewerGui( mainFrame, nes.getMainMemory() );
  
  /** 
   * Creates and configures a new MonkeyNesGui.
   * Note that the window is constructed but not displayed.
   */  
  public MonkeyNesGui()
  {
    // Put the menu system together -------------
    // File
    fileMenu.setMnemonic( KeyEvent.VK_F );
    fileMenu.add( fileOpenMenuItem );
    fileMenu.add( fileCloseMenuItem );
    fileMenu.add( fileExitMenuItem );

    // CPU
    cpuMenu.setMnemonic( KeyEvent.VK_C );
    cpuMenu.add( cpuPowerMenuItem );
    cpuMenu.add( cpuResetMenuItem );

    // Options
    optMenu.setMnemonic( KeyEvent.VK_O );
    optMenu.add( optCpuMenuItem );
    optMenu.add( optVideoMenuItem );
    optMenu.add( optSoundMenuItem );
    optMenu.add( optInputMenuItem );
    optMenu.add( optGameGenieMenuItem );
    
    // Tools
    toolMenu.add( toolCpuStateViewerMenuItem );
    toolMenu.add( toolOpCodeTestShellMenuItem );
    toolMenu.add( toolPatternViewerMenuItem );
    toolMenu.add( toolPrgRomViewerMenuItem );
    toolMenu.add( toolPrgRomHexViewerMenuItem );
    toolMenu.add( toolLoadSmbMenuItem );

    // Help
    helpMenu.setMnemonic( KeyEvent.VK_H );
    helpMenu.add( helpAboutMenuItem );

    // create the actual menu bar
    mainMenu.add( fileMenu );
    mainMenu.add( cpuMenu );
    mainMenu.add( optMenu );
    mainMenu.add( toolMenu );
    mainMenu.add( helpMenu );
    // End Menu System --------------------------

    // Configure the FileChooser
    fileChooser = new JFileChooser();
    fileChooser.setMultiSelectionEnabled( false );
    fileChooser.addChoosableFileFilter( romTypeFilter );

    // Setup the lights
    lightPanel.add( romLoadedLight );
    lightPanel.add( powerOnLight );

    // Build the status label
    statusPanel.add( statusLabel,   BorderLayout.CENTER );
    statusPanel.add( lightPanel,    BorderLayout.EAST );

    // Build the main view
    mainFrame.getContentPane().add( mainMenu,     BorderLayout.NORTH );
    mainFrame.getContentPane().add( statusPanel,  BorderLayout.SOUTH );
    mainFrame.getContentPane().add( outputScreen, BorderLayout.CENTER );
    
    // Hook all the hardware together appropriately
    hookupHardware();

    // ----------------------------------------------------------------------------
    // ----------------------------------------------------------------------------
    // ADD Listeners
    // ----------------------------------------------------------------------------
    // ----------------------------------------------------------------------------

    // Main window closing listener
    mainFrame.addWindowListener( new WindowAdapter()
    {
      public void windowClosing( WindowEvent we )
      {
        shutDown();
      }
    } );
    
    // Key listener for the Joypads
    mainFrame.addKeyListener( padListener );

    // File->Open menu listener
    fileOpenMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        openRom();
      }
    } );

    // File->Close menu listener
    fileCloseMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        closeRom();
      }
    } );

    // File->Exit menu listener
    fileExitMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        shutDown();
      }
    } );

    // CPU->Power menu listener
    cpuPowerMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        if( cpuPowerMenuItem.getState() )
          powerOn();
        else
          powerOff();
      }
    } );

    // CPU->Reset menu listener
    cpuResetMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        DebugLog.log( "CPU->Reset - Not Implemented" );
      }
    } );
    
    // Options->CPU... menu listener
    optCpuMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        optionsCpuWindow.show();
      }
    } );

    // Options->Video... menu listener
    optVideoMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        optionsVideoWindow.show();
      }
    } );


    // Options->Sound... menu listener
    optSoundMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        optionsSoundWindow.show();
      }
    } );

    // Options->Input... menu listener
    optInputMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        optionsInputWindow.show();
        
        // Create a new NesControllerKeyListener with the new keys
        padListener = new NesControllerKeyListener( pad1, pad2 );
      }
    } );
    
    // Options->Game Genie... menu listener
    optGameGenieMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        gameGenieWindow.show();
      }
    } );
    
    // Tools->CPU State Viewer menu listener
    toolCpuStateViewerMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        cpuStateViewerWindow.show();
      }
    } );
    
    // Tools->OpCode Test Shell menu listener
    toolOpCodeTestShellMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        opCodeTestShellWindow.show();
      }
    } );
    
    // Tools->Pattern Table Viewer menu listener
    toolPatternViewerMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        patternTableViewerWindow.show();
      }
    } );
    
    // Tools->PRG-ROM Disassembly Viewer menu listener
    toolPrgRomViewerMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        prgRomViewerWindow.show();
      }
    } );
    
    // Tools->PRG-ROM Hes Viewer menu listener
    toolPrgRomHexViewerMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        prgRomHexViewerWindow.show();
      }
    } );
    
    // Tools->Quick Load SMB
    toolLoadSmbMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        if( rom != null )
          closeRom();      
    
        rom = NesRomLoader.loadRomFile( "D:\\JavaCode\\MonkeyNes\\roms\\SuperMarioBros.nes" );
        //rom = NesRomLoader.loadRomFile( "F:\\MonkeyNes\\roms\\SuperMarioBros.nes" );
        
        nes.insertRom( rom );
        statusLabel.setText( "ROM Loaded" );
        romLoadedLight.setState( true );
        
        adjustMenus();
      }
    } );

    // Help->About menu listener
    helpAboutMenuItem.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        helpAboutWindow.show();
      }
    } );

  } // End of constructor ----------------------------------------------------------

  // Helper methods
  private void shutDown()
  {
    // For now just can everything...
    System.exit(0);
  }
  
  private void hookupHardware()
  {
    nes.connectVideoScreen( outputScreen );
    nes.connectController( pad1, 0 );
    nes.connectController( pad2, 1 );
  }

  private void openRom()
  {
    if( rom != null )
      closeRom();      
    
    fileChooser.setFileSelectionMode( JFileChooser.FILES_AND_DIRECTORIES );
    if( fileChooser.showOpenDialog( mainFrame ) == JFileChooser.APPROVE_OPTION )
    {
      rom = NesRomLoader.loadRomFile( fileChooser.getSelectedFile().getPath() );

      nes.insertRom( rom );
      statusLabel.setText( "ROM Loaded" );
      romLoadedLight.setState( true );
    
      adjustMenus();
    }
  }

  private void closeRom()
  {
    rom = null;
    powerOff();
    nes.removeRom();
    statusLabel.setText( "Idle" );
    romLoadedLight.setState( false );
    powerOnLight.setState( false );
    
    adjustMenus();
  }
  
  private void powerOn()
  {
    powerOnLight.setState( true );
    statusLabel.setText( "Emulating" );
    
    padListener.setPowerOn( true );
    nes.setClockSpeed( optionsCpuWindow.getClockSpeed() );
    nes.powerOn();
    
    DebugLog.log( nes.toString() );
    
    Thread t = new Thread( nes );
    t.start();
    
    adjustMenus();
  }  
  
  private void powerOff()
  {
    powerOnLight.setState( false );
    statusLabel.setText( "ROM Loaded" );
    
    padListener.setPowerOn( false );
    nes.powerOff();
    
    adjustMenus();
  }  
  
  private void adjustMenus()
  {
    // If we are emulating, nothing related to options can be changed
    // If a rom is not loaded, emulation can not be started.

    if( rom == null )
    {
      fileMenu.setEnabled( true );
      fileOpenMenuItem.setEnabled( true );
      fileCloseMenuItem.setEnabled( true );
      fileExitMenuItem.setEnabled( true );
      cpuMenu.setEnabled( false );
      cpuPowerMenuItem.setEnabled( false );
      cpuPowerMenuItem.setState( false );            // State not enabled!
      cpuResetMenuItem.setEnabled( false );
      optMenu.setEnabled( true );
      optCpuMenuItem.setEnabled( true );
      optVideoMenuItem.setEnabled( true );
      optSoundMenuItem.setEnabled( true );
      optInputMenuItem.setEnabled( true );
      optGameGenieMenuItem.setEnabled( true );
      toolMenu.setEnabled( true );
      toolCpuStateViewerMenuItem.setEnabled( true );
      toolOpCodeTestShellMenuItem.setEnabled( true );
      toolPatternViewerMenuItem.setEnabled( true );
      toolPrgRomViewerMenuItem.setEnabled( true );
      toolPrgRomHexViewerMenuItem.setEnabled( true );
      toolLoadSmbMenuItem.setEnabled( true );
      helpMenu.setEnabled( true );
      helpAboutMenuItem.setEnabled( true );
    }
    else
    {
      if( nes.isPowerOn() )
      {
        fileMenu.setEnabled( true );
        fileOpenMenuItem.setEnabled( true );
        fileCloseMenuItem.setEnabled( true );
        fileExitMenuItem.setEnabled( true );
        cpuMenu.setEnabled( true );
        cpuPowerMenuItem.setEnabled( true );
        cpuResetMenuItem.setEnabled( true );
        optMenu.setEnabled( false );
        optCpuMenuItem.setEnabled( false );
        optVideoMenuItem.setEnabled( false );
        optSoundMenuItem.setEnabled( false );
        optInputMenuItem.setEnabled( false );
        optGameGenieMenuItem.setEnabled( false );
        toolMenu.setEnabled( true );
        toolCpuStateViewerMenuItem.setEnabled( true );
        toolOpCodeTestShellMenuItem.setEnabled( true );
        toolPatternViewerMenuItem.setEnabled( true );
        toolPrgRomViewerMenuItem.setEnabled( true );
        toolPrgRomHexViewerMenuItem.setEnabled( true );
        toolLoadSmbMenuItem.setEnabled( true );
        helpMenu.setEnabled( true );
        helpAboutMenuItem.setEnabled( true );
      }  
      else
      {
        fileMenu.setEnabled( true );
        fileOpenMenuItem.setEnabled( true );
        fileCloseMenuItem.setEnabled( true );
        fileExitMenuItem.setEnabled( true );
        cpuMenu.setEnabled( true );
        cpuPowerMenuItem.setEnabled( true );
        cpuResetMenuItem.setEnabled( true );
        optMenu.setEnabled( true );
        optCpuMenuItem.setEnabled( true );
        optVideoMenuItem.setEnabled( true );
        optSoundMenuItem.setEnabled( true );
        optInputMenuItem.setEnabled( true );
        optGameGenieMenuItem.setEnabled( true );
        toolMenu.setEnabled( true );
        toolCpuStateViewerMenuItem.setEnabled( true );
        toolOpCodeTestShellMenuItem.setEnabled( true );
        toolPatternViewerMenuItem.setEnabled( true );
        toolPrgRomViewerMenuItem.setEnabled( true );
        toolPrgRomHexViewerMenuItem.setEnabled( true );
        toolLoadSmbMenuItem.setEnabled( true );
        helpMenu.setEnabled( true );
        helpAboutMenuItem.setEnabled( true );
      }
    }
  }

  /**
   * Displays the MonkeyNES Main window.
   */ 
  public void show()
  {
    // Position the window
    mainFrame.setLocation( new Point( 0, 0 ) );

    // Size the window
    mainFrame.setSize( 400, 400 );

    // Set the title
    mainFrame.setTitle( "MonkeyNES" );

    // Show the whole mess...
    mainFrame.setVisible( true );
    
    adjustMenus();
  }

}