package com.monkeycoder.monkeynes.gui;

import com.monkeycoder.monkeynes.base.*;
import com.monkeycoder.monkeynes.hardware.*;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;

/**
 * This is an interactive GUI dialog that allows for OpCode validation.  This is done through
 * direct manipulation of memory and direct access to the CPU.
 * <br><br>
 * Originally part of the MonkeyNES project.
 *
 * @author David Stephens
 */
public class OpCodeTestShellGui
{
  private static final int OPERATOR_EQUALS         = 0;
  private static final int OPERATOR_NOT_EQUALS     = 1;
  private static final int OPERATOR_LESS_THAN      = 2;
  private static final int OPERATOR_GREATER_THAN   = 3;
  private static final int OPERATOR_LESS_EQUALS    = 4;
  private static final int OPERATOR_GREATER_EQUALS = 5;
  
  private Frame              parentFrame                      = null;
  private JDialog            mainDialog                       = null;
  private boolean            visible                          = false;

  private NesHardware        nes                              = null;

  private Document    outputDoc       = new DefaultStyledDocument();
  private JTextArea   outputArea      = new JTextArea( outputDoc, "", 15, 30 );
  private JScrollPane outputScroll    = new JScrollPane( outputArea );

  private JPanel      bottomPanel     = new JPanel( new BorderLayout() );
  private JTextField  inputText       = new JTextField( "", 200 );

  private ArrayList   cmdHistory      = new ArrayList( 10 );
  private int         histIndex       = -1;

  /**
   * Creates and configures a new OpCodeTestShellGui.
   * Note that the window is constructed but not displayed.
   * @param inParentFrame The parent of this dialog.
   * @param inHardware The NesHardware that the test shell will use.
   */
  public OpCodeTestShellGui( Frame inParentFrame, NesHardware inHardware )
  {
    parentFrame = inParentFrame;
    nes         = inHardware;

    //mainMem.addObserver( this );

    // output area
    outputArea.setFont( new Font( "Courier", Font.PLAIN, 12 ) );
    outputArea.setEditable( false );

    // bottom panel
    bottomPanel.add( new JLabel( ">" ), BorderLayout.WEST   );
    bottomPanel.add( inputText,         BorderLayout.CENTER );

    // Dialog
    mainDialog = new JDialog( parentFrame, "OpCode Test Shell", false );

    // Build the main view
    mainDialog.getContentPane().add( outputScroll,    BorderLayout.CENTER  );
    mainDialog.getContentPane().add( bottomPanel,     BorderLayout.SOUTH  );


    // Input field listener (for pressing enter)
    inputText.addActionListener( new ActionListener()
    {
      public void actionPerformed( ActionEvent ae )
      {
        String command = inputText.getText();
        inputText.setText( "" );

        commandProcessor( command, 0 );
        cmdHistory.add( new String( command ) );
        histIndex = -1;

        outputArea.setCaretPosition( outputArea.getText().length() );
        inputText.requestFocus();
      }
    } );

    // Input field listener (for pressing enter)
    inputText.addKeyListener( new KeyAdapter()
    {
      public void keyPressed( KeyEvent ke )
      {
        boolean controlKey = false;

        // Up Arrow
        if( ke.getKeyCode() == 38 )
        {
          if( histIndex == -1 )
            histIndex = cmdHistory.size() - 1;
          else if( histIndex > 0 )
            histIndex--;

          controlKey = true;
        }
        // Down Arrow
        else if( ke.getKeyCode() == 40 )
        {
          if( histIndex != -1 && histIndex < (cmdHistory.size() - 1) )
            histIndex++;
          else
            histIndex = -1;

          controlKey = true;
        }

        if( controlKey )
        {
          if( histIndex == -1 )
            inputText.setText( "" );
          else
            inputText.setText( (String)cmdHistory.get( histIndex ) );
        }
      }
    } );

    // Main window closing listener
    mainDialog.addWindowListener( new WindowAdapter()
    {
      public void windowClosing( WindowEvent we )
      {
        mainDialog.hide();
        visible = false;
      }
    } );
  } // End of constructor ----------------------------------------------------------


  /**
   * Takes a command and does the requested action.
   * @param inCommand The command to process.
   * @param depth The call depth of this processor for resursive activities.
   */
  private void commandProcessor( String inCommand, int depth )
  {
    StringTokenizer cmdTok  = new StringTokenizer( inCommand );
    String          keyTok  = "";
    String          subTok1 = "";
    String          subTok2 = "";
    String          subTok3 = "";
    int             cycles  = 0;

    if( cmdTok.hasMoreTokens() )
    {
      outputArea.append( "> " + inCommand + "\n" );

      keyTok = cmdTok.nextToken();

      // ------------------------------------------------------------------------------------------
      if( keyTok.equalsIgnoreCase( "help" ) )
      {
        if( cmdTok.hasMoreTokens() )
        {
          subTok1 = cmdTok.nextToken();

          if( subTok1.equalsIgnoreCase( "clear" ) )
          {
            outputArea.append( "Syntax:  clear\n" );
            outputArea.append( "  Clears the output buffer.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "#" ) )
          {
            outputArea.append( "Syntax:  # [comment]\n" );
            outputArea.append( "  comment - Text that will not be parsed by the command processor.\n" );
            outputArea.append( "            This text will still be outputted to the shell.\n" );
            outputArea.append( "- Note that the space following the # is required as the # is \n" );
            outputArea.append( "  actually the command.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "help" ) )
          {
            outputArea.append( "Syntax:  help [command]\n" );
            outputArea.append( "  command - The command for which you want more detailed help.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "load" ) )
          {
            outputArea.append( "Syntax:  load [script]\n" );
            outputArea.append( "  script - The path to a script file containing a sequence of\n" );
            outputArea.append( "           test commands to execute.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "memclear" ) )
          {
            outputArea.append( "Syntax:  memclear\n" );
            outputArea.append( "  Clears main memory to all zeros.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "memget" ) )
          {
            outputArea.append( "Syntax:  memget [value]\n" );
            outputArea.append( "  addr  - 2 byte hex position in memory to get the value from.\n" );
            outputArea.append( "- Accesses main system memory.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "memset" ) )
          {
            outputArea.append( "Syntax:  memset [addr] [value]\n" );
            outputArea.append( "  addr  - 2 byte hex position in memory to set.\n" );
            outputArea.append( "  value - 1 byte hex value to set.\n" );
            outputArea.append( "- Accesses main system memory.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "op" ) )
          {
            outputArea.append( "Syntax:  op [opcode] [param]\n" );
            outputArea.append( "  opcode - The hex of the opcode to execute.\n" );
            outputArea.append( "  param  - The 0 to 2 bytes of hex to use as the params to the\n" );
            outputArea.append( "           opcode.\n" );
            outputArea.append( "- For more info about opcodes and their expected params please\n" );
            outputArea.append( "  check out http://monkeycoder.com/monkeynes/opcodes.php\n" );
            outputArea.append( "- All hex is expected as plain text with no prefix of '$' or '0x'\n" );
            outputArea.append( "  and is expected to be 0 padded appropriately.\n" );
            outputArea.append( "- Please note param values are expected low order byte first like\n" );
            outputArea.append( "  they would be from the rom.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "opx" ) )
          {
            outputArea.append( "Syntax:  opx [6502]\n" );
            outputArea.append( "  6502 - Operation to execute in 6502 assembly.\n" );
            outputArea.append( "- For more info about opcode syntax please check out\n" );
            outputArea.append( "  http://monkeycoder.com/monkeynes/opcodes.php\n" );
            outputArea.append( "- This command parses the 6502 code and passes it to\n" );
            outputArea.append( "  the op command.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "power" ) )
          {
            outputArea.append( "Syntax:  power [state]\n" );
            outputArea.append( "  state - on|off - turns on or off the NES hardware.\n" );
            outputArea.append( "                   (just sets register values)\n" );
          }
          else if( subTok1.equalsIgnoreCase( "regs" ) )
          {
            outputArea.append( "Syntax:  regs [reg]\n" );
            outputArea.append( "  reg - Optionally supply the register to display the value of.\n" );
            outputArea.append( "        Not supplying this parameter will display all regs.\n" );
            outputArea.append( "- This will output the current state of all the registers in\n" );
            outputArea.append( "  the CPU.  Very similar to the CPU State Viewer output except\n" );
            outputArea.append( "  that it's in text.\n\n" );
            outputArea.append( "Supported register names are:\n" );
            outputArea.append( "  PC     - Program Counter Register\n" );
            outputArea.append( "  SP     - Stack Pointer Register\n" );
            outputArea.append( "  ACC    - Accumulator Register\n" );
            outputArea.append( "  IX     - X-Index Register\n" );
            outputArea.append( "  IY     - Y-Index Register\n" );
            outputArea.append( "  PS     - Processor Status Register (as a whole)\n" );
            outputArea.append( "  S      - Negative Flag\n" );
            outputArea.append( "  V      - Overflow Flag\n" );
            outputArea.append( "  B      - Break Command Flag\n" );
            outputArea.append( "  D      - Decimal Mode Flag\n" );
            outputArea.append( "  I      - Interrupt Disable Flag\n" );
            outputArea.append( "  Z      - Zero Flag\n" );
            outputArea.append( "  C      - Carry Flag\n" );
          }
          else if( subTok1.equalsIgnoreCase( "regset" ) )
          {
            outputArea.append( "Syntax:  regs [reg] [value]\n" );
            outputArea.append( "  reg   - The register to set the value of.\n" );
            outputArea.append( "  value - The value the register will be set to, in hex.\n\n" );
            outputArea.append( "Supported register names are:\n" );
            outputArea.append( "  PC     - Program Counter Register\n" );
            outputArea.append( "  SP     - Stack Pointer Register\n" );
            outputArea.append( "  ACC    - Accumulator Register\n" );
            outputArea.append( "  IX     - X-Index Register\n" );
            outputArea.append( "  IY     - Y-Index Register\n" );
            outputArea.append( "  PS     - Processor Status Register (as a whole)\n" );
            outputArea.append( "  S      - Negative Flag\n" );
            outputArea.append( "  V      - Overflow Flag\n" );
            outputArea.append( "  B      - Break Command Flag\n" );
            outputArea.append( "  D      - Decimal Mode Flag\n" );
            outputArea.append( "  I      - Interrupt Disable Flag\n" );
            outputArea.append( "  Z      - Zero Flag\n" );
            outputArea.append( "  C      - Carry Flag\n" );
          }
          else if( subTok1.equalsIgnoreCase( "rom" ) )
          {
            outputArea.append( "Syntax:  rom [file]\n" );
            outputArea.append( "  file - The path to a *.NES ROM file to load.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "save" ) )
          {
            outputArea.append( "Syntax:  save [file]\n" );
            outputArea.append( "  file - The path to a file to which the shell output will\n" );
            outputArea.append( "         be saved.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "step" ) )
          {
            outputArea.append( "Syntax:  step\n" );
            outputArea.append( "  This function executes the next instruction from memory based\n" );
            outputArea.append( "  on the current CPU state.  The number of cycles required to\n" );            
            outputArea.append( "  execute the instruction are automatically granted.\n" );            
          }
          else if( subTok1.equalsIgnoreCase( "test" ) )
          {
            outputArea.append( "Syntax:  test [condition]\n" );
            outputArea.append( "  condition - This is the comparison to perform.  The result must\n" );
            outputArea.append( "              be boolean and a true or false will be displayed.\n" );
            outputArea.append( "              The condition is made up of three pieces, and is\n" );
            outputArea.append( "              always in the form: [variable] [operator] [value]\n" );
            outputArea.append( "  variable  - The variable is some changing part of the system\n" );
            outputArea.append( "              that holds a value you want to compare against.\n" );
            outputArea.append( "  operator  - The comparison operator to use.  Supported operators\n" );
            outputArea.append( "              are:  =, !=, <, >, <=, >=\n" );
            outputArea.append( "  value     - Any appropriately sized numeric value, always in hex.\n\n" );
            outputArea.append( "Supported variables are:\n" );
            outputArea.append( "  PC     - Program Counter Register\n" );
            outputArea.append( "  SP     - Stack Pointer Register\n" );
            outputArea.append( "  ACC    - Accumulator Register\n" );
            outputArea.append( "  IX     - X-Index Register\n" );
            outputArea.append( "  IY     - Y-Index Register\n" );
            outputArea.append( "  PS     - Processor Status Register (as a whole)\n" );
            outputArea.append( "  S      - Negative Flag\n" );
            outputArea.append( "  V      - Overflow Flag\n" );
            outputArea.append( "  B      - Break Command Flag\n" );
            outputArea.append( "  D      - Decimal Mode Flag\n" );
            outputArea.append( "  I      - Interrupt Disable Flag\n" );
            outputArea.append( "  Z      - Zero Flag\n" );
            outputArea.append( "  C      - Carry Flag\n" );
            outputArea.append( "  m:XXXX - Value at address XXXX (hex) in main memory.\n" );
            outputArea.append( "  v:XXXX - Value at address XXXX (hex) in video memory.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "vidclear" ) )
          {
            outputArea.append( "Syntax:  vidclear\n" );
            outputArea.append( "  Clears video memory to all zeros.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "vidget" ) )
          {
            outputArea.append( "Syntax:  vidget [value]\n" );
            outputArea.append( "  addr  - 2 byte hex position in memory to get the value from.\n" );
            outputArea.append( "- Accesses video memory.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "vidset" ) )
          {
            outputArea.append( "Syntax:  vidget [addr] [value]\n" );
            outputArea.append( "  addr  - 2 byte hex position in memory to set.\n" );
            outputArea.append( "  value - 1 byte hex value to set.\n" );
            outputArea.append( "- Accesses video memory.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "monkey" ) )
          {
            outputArea.append( "Help monkey?  Aren't I already?\n" );
          }
          else
          {
            outputArea.append( "No help available for command: '" + subTok1 + "'\n" );
          }
        }
        else
        {
          outputArea.append( "Syntax: help [command]\n" );
          outputArea.append( "  command - the command for which you want more detailed help.\n\n" );
          outputArea.append( "Command List:\n" );
          outputArea.append( "  #         clear     help      load      memclear  \n" );
          outputArea.append( "  memget    memset    op        opx       power     \n" );
          outputArea.append( "  regs      regset    rom       save      step      \n" );
          outputArea.append( "  test      vidclear  vidget    vidset    \n" );
        }
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "#" ) )
      {
        // The output for this command already displayed on the prompt line.
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "clear" ) )
      {
        if( depth < 2 )
          initialize();
        else
          outputArea.append( "Clear command disabled within nested scripts.\n" );
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "load" ) )
      {
        if( depth < 10 )
        {
          if( cmdTok.hasMoreTokens() )
          {
            subTok1 = inCommand.substring( 5 );

            try
            {
              LineNumberReader lnr = new LineNumberReader( new FileReader( subTok1 ) );
              String           cur = "";

              outputArea.append( "Executing Script: '" + subTok1 + "'.\n\n" );
              while( (cur = lnr.readLine()) != null )
              {
                commandProcessor( cur, (depth + 1) );
                outputArea.setCaretPosition( outputArea.getText().length() );
              }
              outputArea.append( "Script Complete.\n" );

              lnr.close();
            }
            catch( IOException ioe )
            {
              outputArea.append( "Failed to Access Script: '" + subTok1 + "'.\n\n" );
              return;
            }
          }
          else
          {
            outputArea.append( "Expected Script File Name.\n" );
          }
        }
        else
        {
          outputArea.append( "Max of 10 layers of Nested File Load Depth Exceeded.\n" );
          outputArea.append( "(This is to prevent cyclic file loading.)\n" );
        }
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "memclear" ) )
      {
        nes.getMainMemory().clear();
        outputArea.append( "Main memory cleared.\n" );
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "memget" ) )
      {
        int param1 = 0;

        // Load the params
        subTok1 = null;
        if( cmdTok.hasMoreTokens() )
          subTok1 = cmdTok.nextToken();
        try
        {
          param1 = Integer.parseInt( subTok1, 16 );
        }
        catch( NumberFormatException nfe )
        {
          outputArea.append( "Address must be numeric.\n\n" );
          return;
        }

        String hexVal = Integer.toString( nes.getMainMemory().readNoMessage( param1 ), 16 );
        if( hexVal.length() == 1 )
          hexVal = "0" + hexVal;
        outputArea.append( "Main Memory[$" + subTok1.toUpperCase() + "]:  $" + hexVal.toUpperCase() + "\n" );
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "memset" ) )
      {
        int param1 = 0;
        int param2 = 0;

        // Load the params
        subTok1 = null;
        subTok2 = null;
        if( cmdTok.hasMoreTokens() )
          subTok1 = cmdTok.nextToken();
        if( cmdTok.hasMoreTokens() )
          subTok2 = cmdTok.nextToken();
        try
        {
          param1 = Integer.parseInt( subTok1, 16 );
          param2 = Integer.parseInt( subTok2, 16 );
        }
        catch( NumberFormatException nfe )
        {
          outputArea.append( "Address and Value must be numeric.\n\n" );
          return;
        }

        nes.getMainMemory().writeNoMessage( param1, param2 );

        String hexVal = Integer.toString( nes.getMainMemory().readNoMessage( param1 ), 16 );
        if( hexVal.length() == 1 )
          hexVal = "0" + hexVal;
        outputArea.append( "Main Memory[$" + subTok1.toUpperCase() + "] set to $" + hexVal.toUpperCase() + "\n" );
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "power" ) )
      {
        if( cmdTok.hasMoreTokens() )
        {
          subTok1 = cmdTok.nextToken();

          if( subTok1.equalsIgnoreCase( "on" ) )
          {
            nes.powerOn();
            outputArea.append( "Power-On settings loaded.\n" );
          }
          else if( subTok1.equalsIgnoreCase( "off" ) )
          {
            nes.getCpu().powerOff();
            outputArea.append( "CPU Power-Off settings loaded.\n" );
          }
          else
          {
            outputArea.append( "Unknown State: '" + subTok1 + "'\n" );
          }
        }
        else
        {
          outputArea.append( "Unknown State: ''\n" );
        }
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "regs" ) )
      {
        // Just in case I need it (since most displays do)
        CpuRegister ps = new CpuRegister( nes.getCpu().getProcessorStatusValue() );
        
        if( cmdTok.hasMoreTokens() )
        {
          subTok1 = cmdTok.nextToken();
          
          if( subTok1.equalsIgnoreCase( "pc" ) )
            outputArea.append( "  PC:    $" + Integer.toString( nes.getCpu().getProgramCounterValue(), 16 ).toUpperCase() + "\n" );
          else if( subTok1.equalsIgnoreCase( "sp" ) )
            outputArea.append( "  SP:    $" + Integer.toString( nes.getCpu().getStackPointerValue(), 16 ).toUpperCase() + "\n" );  
          else if( subTok1.equalsIgnoreCase( "acc" ) )
            outputArea.append( "  ACC:   $" + Integer.toString( nes.getCpu().getAccumulatorValue(), 16 ).toUpperCase() + "\n" );
          else if( subTok1.equalsIgnoreCase( "ix" ) )
            outputArea.append( "  IX:    $" + Integer.toString( nes.getCpu().getXIndexValue(), 16 ).toUpperCase() + "\n" );
          else if( subTok1.equalsIgnoreCase( "iy" ) )
            outputArea.append( "  IY:    $" + Integer.toString( nes.getCpu().getYIndexValue(), 16 ).toUpperCase() + "\n" );
          else if( subTok1.equalsIgnoreCase( "ps" ) )
            outputArea.append( "  PS:    $" + Integer.toString( nes.getCpu().getProcessorStatusValue(), 16 ).toUpperCase() + "\n" );
          else if( subTok1.equalsIgnoreCase( "s" ) )
            outputArea.append( "  S:      " + String.valueOf( ps.getBitValue( CpuRegister.PS_NEGATIVE_FLAG ) ) + "\n" );
          else if( subTok1.equalsIgnoreCase( "v" ) )
            outputArea.append( "  V:      " + String.valueOf( ps.getBitValue( CpuRegister.PS_OVERFLOW_FLAG ) ) + "\n" );
          else if( subTok1.equalsIgnoreCase( "b" ) )
            outputArea.append( "  B:      " + String.valueOf( ps.getBitValue( CpuRegister.PS_BREAK_COMMAND ) ) + "\n" );
          else if( subTok1.equalsIgnoreCase( "d" ) )
            outputArea.append( "  D:      " + String.valueOf( ps.getBitValue( CpuRegister.PS_DECIMAL_MODE ) ) + "\n" );
          else if( subTok1.equalsIgnoreCase( "i" ) )  
            outputArea.append( "  I:      " + String.valueOf( ps.getBitValue( CpuRegister.PS_INTERRUPT_DISABLE ) ) + "\n" );
          else if( subTok1.equalsIgnoreCase( "z" ) )  
            outputArea.append( "  Z:      " + String.valueOf( ps.getBitValue( CpuRegister.PS_ZERO_FLAG ) ) + "\n" );
          else if( subTok1.equalsIgnoreCase( "c" ) )  
            outputArea.append( "  C:      " + String.valueOf( ps.getBitValue( CpuRegister.PS_CARRY_FLAG ) ) + "\n" );
          else
            outputArea.append( "Unknown Register: " + subTok1 + "\n" );          
        }  
        else
        {  
          outputArea.append( "CPU Registers\n" );
          outputArea.append( "  PC:    $" + Integer.toString( nes.getCpu().getProgramCounterValue(), 16 ).toUpperCase() + "\n" );
          outputArea.append( "  SP:    $" + Integer.toString( nes.getCpu().getStackPointerValue(), 16 ).toUpperCase() + "\n" );
          outputArea.append( "  ACC:   $" + Integer.toString( nes.getCpu().getAccumulatorValue(), 16 ).toUpperCase() + "\n" );
          outputArea.append( "  IX:    $" + Integer.toString( nes.getCpu().getXIndexValue(), 16 ).toUpperCase() + "\n" );
          outputArea.append( "  IY:    $" + Integer.toString( nes.getCpu().getYIndexValue(), 16 ).toUpperCase() + "\n" );
          outputArea.append( "  Flags  S V - B D I Z C\n" );
          outputArea.append( "  PS:    " );
          outputArea.append( String.valueOf( ps.getBitValue( CpuRegister.PS_NEGATIVE_FLAG     ) ) + " " );
          outputArea.append( String.valueOf( ps.getBitValue( CpuRegister.PS_OVERFLOW_FLAG     ) ) + " " );
          outputArea.append( String.valueOf( ps.getBitValue( CpuRegister.PS_UNUSED            ) ) + " " );
          outputArea.append( String.valueOf( ps.getBitValue( CpuRegister.PS_BREAK_COMMAND     ) ) + " " );
          outputArea.append( String.valueOf( ps.getBitValue( CpuRegister.PS_DECIMAL_MODE      ) ) + " " );
          outputArea.append( String.valueOf( ps.getBitValue( CpuRegister.PS_INTERRUPT_DISABLE ) ) + " " );
          outputArea.append( String.valueOf( ps.getBitValue( CpuRegister.PS_ZERO_FLAG         ) ) + " " );
          outputArea.append( String.valueOf( ps.getBitValue( CpuRegister.PS_CARRY_FLAG        ) ) + "\n" );
        }  
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "regset" ) )
      {
        int     value   = 0;
        boolean flagSet = false;
        
        // Just in case I need it (since most displays do)
        CpuRegister ps = new CpuRegister( nes.getCpu().getProcessorStatusValue() );
        
        // Load the params
        subTok1 = null;
        subTok2 = null;
        if( cmdTok.hasMoreTokens() )
          subTok1 = cmdTok.nextToken();
        if( cmdTok.hasMoreTokens() )
          subTok2 = cmdTok.nextToken();
                
        if( subTok1 != null )
        {
          if( subTok2 != null )
          {
            try
            {
              value = Integer.parseInt( subTok2, 16 );
            }
            catch( NumberFormatException nfe )
            {
              outputArea.append( "Value must be numeric.\n\n" );
              return;
            }
            
            if( subTok1.equalsIgnoreCase( "pc" ) )
            {
              nes.getCpu().setProgramCounterValue( value );
              outputArea.append( "PC set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
            }
            else if( subTok1.equalsIgnoreCase( "sp" ) )
            {
              nes.getCpu().setStackPointerValue( value );
              outputArea.append( "SP set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
            }
            else if( subTok1.equalsIgnoreCase( "acc" ) )
            {
              nes.getCpu().setAccumulatorValue( value );
              outputArea.append( "ACC set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
            }
            else if( subTok1.equalsIgnoreCase( "ix" ) )
            {
              nes.getCpu().setXIndexValue( value );
              outputArea.append( "IX set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
            }
            else if( subTok1.equalsIgnoreCase( "iy" ) )
            {
              nes.getCpu().setYIndexValue( value );
              outputArea.append( "IY set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
            }
            else if( subTok1.equalsIgnoreCase( "ps" ) )
            {
              nes.getCpu().setProcessorStatusValue( value );
              outputArea.append( "PS set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
            }
            else if( subTok1.equalsIgnoreCase( "s" ) )
            {
              ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, value );
              outputArea.append( "S set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
              flagSet = true;
            }  
            else if( subTok1.equalsIgnoreCase( "v" ) )
            {
              ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, value );
              outputArea.append( "V set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
              flagSet = true;
            } 
            else if( subTok1.equalsIgnoreCase( "b" ) )
            {
              ps.setBitValue( CpuRegister.PS_BREAK_COMMAND, value );
              outputArea.append( "B set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
              flagSet = true;
            } 
            else if( subTok1.equalsIgnoreCase( "d" ) )
            {
              ps.setBitValue( CpuRegister.PS_DECIMAL_MODE, value );
              outputArea.append( "D set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
              flagSet = true;
            } 
            else if( subTok1.equalsIgnoreCase( "i" ) )  
            {
              ps.setBitValue( CpuRegister.PS_INTERRUPT_DISABLE, value );
              outputArea.append( "I set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
              flagSet = true;
            } 
            else if( subTok1.equalsIgnoreCase( "z" ) )  
            {
              ps.setBitValue( CpuRegister.PS_ZERO_FLAG, value );
              outputArea.append( "Z set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
              flagSet = true;
            } 
            else if( subTok1.equalsIgnoreCase( "c" ) )  
            {
              ps.setBitValue( CpuRegister.PS_CARRY_FLAG, value );
              outputArea.append( "C set to $" + Integer.toString( value, 16 ).toUpperCase() + "\n" );
              flagSet = true;
            } 
            else
              outputArea.append( "Unknown Register: " + subTok1 + "\n" );          
              
            // Reload the PS register if a flag was set
            if( flagSet )
              nes.getCpu().setProcessorStatusValue( ps.getValue() );
          }
          else
            outputArea.append( "Expected Value.\n" );          
        }  
        else
          outputArea.append( "Expected Register Name.\n" );          
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "rom" ) )
      {
        if( cmdTok.hasMoreTokens() )
        {
          subTok1 = inCommand.substring( 4 );
 
          NesRom rom = NesRomLoader.loadRomFile( subTok1 );

          if( rom != null )
          {
            nes.insertRom( rom );
            outputArea.append( "ROM Loaded and Inserted into Hardware.\n" );
          }  
          else
            outputArea.append( "Failed to load ROM File: '" + subTok1 + "'.\n" );
        }
        else
        {
          outputArea.append( "Expected ROM File Name.\n" );
        }
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "op" ) )
      {
        int opCode = -1;
        int param1 = 0;
        int param2 = 0;

        // Load the params
        subTok1 = null;
        subTok2 = null;
        if( cmdTok.hasMoreTokens() )
          subTok1 = cmdTok.nextToken();
        if( cmdTok.hasMoreTokens() )
          subTok2 = cmdTok.nextToken();

        if( subTok1 != null )
        {
          try
          {
            opCode = Integer.parseInt( subTok1, 16 );
          }
          catch( NumberFormatException nfe )
          {
            outputArea.append( "OpCode was not numeric.\n\n" );
            return;
          }

          if( NesOpCodeTable.length[opCode] == 1 )
          {
            // opcodes with no parameters
            outputArea.append( "Syntax Ok: " + NesOpCodeTable.getSyntaxFormatted( opCode, 0, 0 ) + "\n" );
            cycles = nes.getCpu().testExecuteInstruction( opCode, 0, 0 );
            outputArea.append( cycles + " cycles.\n" );
          }
          else if( NesOpCodeTable.length[opCode] > 1 )
          {
            if( subTok2 != null )
            {
              if( NesOpCodeTable.length[opCode] == 2 && subTok2.length() == 2 )
              {
                // opcodes with 1 byte parameters
                try
                {
                  param1 = Integer.parseInt( subTok2, 16 );
                }
                catch( NumberFormatException nfe )
                {
                  outputArea.append( "Params were not numeric.\n\n" );
                  return;
                }

                outputArea.append( "Syntax Ok: " + NesOpCodeTable.getSyntaxFormatted( opCode, param1, 0 ) + "\n" );
                cycles = nes.getCpu().testExecuteInstruction( opCode, param1, 0 );
                outputArea.append( cycles + " cycles.\n" );
              }
              else if( NesOpCodeTable.length[opCode] == 3 && subTok2.length() == 4 )
              {
                // opcodes with 2 byte parameters
                try
                {
                  param1 = Integer.parseInt( subTok2.substring( 0, 2 ), 16 );
                  param2 = Integer.parseInt( subTok2.substring( 2, 4 ), 16 );
                }
                catch( NumberFormatException nfe )
                {
                  outputArea.append( "Params were not numeric.\n\n" );
                  return;
                }

                outputArea.append( "Syntax Ok: " + NesOpCodeTable.getSyntaxFormatted( opCode, param1, param2 ) + "\n" );
                cycles = nes.getCpu().testExecuteInstruction( opCode, param1, param2 );
                outputArea.append( cycles + " cycles.\n" );
              }
              else
              {
                outputArea.append( NesOpCodeTable.name[opCode] + " requires " + (NesOpCodeTable.length[opCode] - 1) + " byte param.\n" );
              }
            }
            else
            {
              outputArea.append( NesOpCodeTable.name[opCode] + " requires " + (NesOpCodeTable.length[opCode] - 1) + " byte param.\n" );
            }
          }
        }
        else
        {
          outputArea.append( "Expected opcode." );
        }
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "opx" ) )
      {
        int opCode = -1;
        String param = "";
        int param1 = 0;
        int param2 = 0;

        if( cmdTok.hasMoreTokens() )
        {
          subTok1 = cmdTok.nextToken();

          if( cmdTok.hasMoreTokens() )
          {
            subTok2 = inCommand.substring( 8 );
          }

          opCode = NesOpCodeTable.getOpCodeBySyntax( subTok1, subTok2 );

          if( opCode != -1 )
          {
            param = NesOpCodeTable.getParam( subTok2 );

            StringBuffer opCommand = new StringBuffer( "op " );
            opCommand.append( Integer.toString( opCode, 16 ).toUpperCase() );
            opCommand.append( " " );
            if( param.length() == 2 )
            {
              opCommand.append( param.substring( 0, 2 ) );
            }
            else if( param.length() == 4 )
            {
              opCommand.append( param.substring( 2, 4 ) );
              opCommand.append( param.substring( 0, 2 ) );
            }

            // This does not increase the depth of nested instructions
            commandProcessor( opCommand.toString(), depth );
            outputArea.setCaretPosition( outputArea.getText().length() );
          }
          else
          {
            outputArea.append( "Unrecognized 6502 syntax.\n" );
          }
        }
        else
        {
          outputArea.append( "Expected 6502 Assembly Operation.\n" );
        }
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "save" ) )
      {
        if( cmdTok.hasMoreTokens() )
        {
          subTok1 = inCommand.substring( 5 );

          try
          {
            FileOutputStream fos = new FileOutputStream( subTok1 );
            fos.write( outputArea.getText().getBytes() );
            fos.flush();
            fos.close();
            outputArea.append( "Shell Output Saved to: '" + subTok1 + "'.\n" );
          }
          catch( IOException ioe )
          {
            outputArea.append( "Failed to Save Shell Output.\n\n" );
            return;
          }
        }
        else
        {
          outputArea.append( "Expected Script File Name.\n" );
        }
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "step" ) )
      {
        ExecutedInstructionInfo eii = nes.getCpu().testStepInstruction();
        outputArea.append( "Executed: " + eii.getSyntaxFormatted() + "\n" );
        outputArea.append( eii.getCycles() + " cycles.\n" );
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "test" ) )
      {
        boolean     valid    = true;                // whether or not all the pieces were supplied correctly
        int         addr     = 0;                   // address portion of memory access variables
        int         variable = 0;                   // holds the actual value of the variable
        int         operator = OPERATOR_EQUALS;     // holds the type of comparison that will be done
        int         value    = 0;                   // holds the value to test against
        CpuRegister ps       = new CpuRegister( nes.getCpu().getProcessorStatusValue() );

        // First the variable
        if( cmdTok.hasMoreTokens() )
        {
          subTok1 = cmdTok.nextToken();

          if( subTok1.equalsIgnoreCase( "PC" ) )
          {
            variable = nes.getCpu().getProgramCounterValue();
          }
          else if( subTok1.equalsIgnoreCase( "SP" ) )
          {
            variable = nes.getCpu().getStackPointerValue();
          }
          else if( subTok1.equalsIgnoreCase( "ACC" ) )
          {
            variable = nes.getCpu().getAccumulatorValue();
          }
          else if( subTok1.equalsIgnoreCase( "IX" ) )
          {
            variable = nes.getCpu().getXIndexValue();
          }
          else if( subTok1.equalsIgnoreCase( "IY" ) )
          {
            variable = nes.getCpu().getYIndexValue();
          }
          else if( subTok1.equalsIgnoreCase( "PS" ) )
          {
            variable = nes.getCpu().getProcessorStatusValue();
          } 
          else if( subTok1.equalsIgnoreCase( "S" ) )
          {
            variable = ps.getBitValue( CpuRegister.PS_NEGATIVE_FLAG );
          }
          else if( subTok1.equalsIgnoreCase( "V" ) )
          {
            variable = ps.getBitValue( CpuRegister.PS_OVERFLOW_FLAG );
          }
          else if( subTok1.equalsIgnoreCase( "B" ) )
          {
            variable = ps.getBitValue( CpuRegister.PS_BREAK_COMMAND );
          }
          else if( subTok1.equalsIgnoreCase( "D" ) )
          {
            variable = ps.getBitValue( CpuRegister.PS_DECIMAL_MODE );
          }
          else if( subTok1.equalsIgnoreCase( "I" ) )
          {
            variable = ps.getBitValue( CpuRegister.PS_INTERRUPT_DISABLE );
          }
          else if( subTok1.equalsIgnoreCase( "Z" ) )
          {
            variable = ps.getBitValue( CpuRegister.PS_ZERO_FLAG );
          }
          else if( subTok1.equalsIgnoreCase( "C" ) )
          {
            variable = ps.getBitValue( CpuRegister.PS_CARRY_FLAG );
          }
          else
          {
            if( subTok1.length() == 6 )
            {
              try
              {
                addr = Integer.parseInt( subTok1.substring( 2, 6 ), 16 );
              }
              catch( NumberFormatException nfe )
              {
                outputArea.append( "Address portion of variable must be numeric.\n" );
                valid = false;
              }
              
              if( addr != 0 )
              {
                if( subTok1.substring( 0, 2 ).equalsIgnoreCase( "m:" ) )
                {
                  variable = nes.getMainMemory().readNoMessage( addr );
                }
                else if( subTok1.substring( 0, 2 ).equalsIgnoreCase( "v:" ) )
                {
                  variable = nes.getVideoMemory().readNoMessage( addr );
                }
                else
                {
                  outputArea.append( "Unknown Variable: '" + subTok1 + "'.\n" );
                  valid = false;
                }
              } 
            }
            else
            {
              outputArea.append( "Unknown Variable: '" + subTok1 + "'.\n" );
              valid = false;
            }
          }
        }
        else
        {
          outputArea.append( "Expected Variable.\n" );
        }

        // Then the operator
        if( cmdTok.hasMoreTokens() )
        {
          subTok2 = cmdTok.nextToken();

          if( subTok2.equalsIgnoreCase( "=" ) )
          {
            operator = OPERATOR_EQUALS;
          } 
          else if( subTok2.equalsIgnoreCase( "!=" ) )
          { 
            operator = OPERATOR_NOT_EQUALS;
          }
          else if( subTok2.equalsIgnoreCase( "<" ) )
          {
            operator = OPERATOR_LESS_THAN; 
          }
          else if( subTok2.equalsIgnoreCase( ">" ) )
          {
            operator = OPERATOR_GREATER_THAN;
          }
          else if( subTok2.equalsIgnoreCase( "<=" ) )
          {
            operator = OPERATOR_LESS_EQUALS; 
          }
          else if( subTok2.equalsIgnoreCase( ">=" ) )
          {
            operator = OPERATOR_GREATER_EQUALS;
          }          
          else
          {
            outputArea.append( "Unknown Operator: '" + subTok2 + "'.\n" );
            valid = false;
          }
        }
        else
        {
          outputArea.append( "Expected Operator.\n" );
          valid = false;
        }

        // And finally the value
        if( cmdTok.hasMoreTokens() )
        {
          subTok3 = cmdTok.nextToken();

          try
          {
            value = Integer.parseInt( subTok3, 16 );
          }
          catch( NumberFormatException nfe )
          {
            outputArea.append( "Value must be numeric.\n" );
            valid = false;
          }
        }
        else
        {
          outputArea.append( "Expected Value.\n" );
          valid = false;
        }

        // Now do the comparison and output true or false
        if( valid )
        {
          
          if( operator == OPERATOR_EQUALS )
          {
            outputArea.append( Boolean.toString( variable == value ) + "\n" );
          }
          else if( operator == OPERATOR_NOT_EQUALS )
          {
            outputArea.append( Boolean.toString( variable != value ) + "\n" );
          }
          else if( operator == OPERATOR_LESS_THAN )
          {
            outputArea.append( Boolean.toString( variable < value ) + "\n" );
          }
          else if( operator == OPERATOR_GREATER_THAN )
          {
            outputArea.append( Boolean.toString( variable > value ) + "\n" );
          }
          else if( operator == OPERATOR_LESS_EQUALS )
          {
            outputArea.append( Boolean.toString( variable <= value ) + "\n" );
          }
          else if( operator == OPERATOR_GREATER_EQUALS )
          {
            outputArea.append( Boolean.toString( variable >= value ) + "\n" );
          }
        }
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "vidclear" ) )
      {
        nes.getVideoMemory().clear();
        outputArea.append( "Video memory cleared.\n" );
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "vidget" ) )
      {
        int param1 = 0;

        // Load the params
        subTok1 = null;
        if( cmdTok.hasMoreTokens() )
          subTok1 = cmdTok.nextToken();
        try
        {
          param1 = Integer.parseInt( subTok1, 16 );
        }
        catch( NumberFormatException nfe )
        {
          outputArea.append( "Address must be numeric.\n\n" );
          return;
        }

        String hexVal = Integer.toString( nes.getVideoMemory().readNoMessage( param1 ), 16 );
        if( hexVal.length() == 1 )
          hexVal = "0" + hexVal;
        outputArea.append( "Video Memory[$" + subTok1.toUpperCase() + "]:  $" + hexVal.toUpperCase() + "\n" );
      }
      // ------------------------------------------------------------------------------------------
      else if( keyTok.equalsIgnoreCase( "vidset" ) )
      {
        int param1 = 0;
        int param2 = 0;

        // Load the params
        subTok1 = null;
        subTok2 = null;
        if( cmdTok.hasMoreTokens() )
          subTok1 = cmdTok.nextToken();
        if( cmdTok.hasMoreTokens() )
          subTok2 = cmdTok.nextToken();
        try
        {
          param1 = Integer.parseInt( subTok1, 16 );
          param2 = Integer.parseInt( subTok2, 16 );
        }
        catch( NumberFormatException nfe )
        {
          outputArea.append( "Address and Value must be numeric.\n\n" );
          return;
        }

        nes.getVideoMemory().writeNoMessage( param1, param2 );

        String hexVal = Integer.toString( nes.getVideoMemory().readNoMessage( param1 ), 16 );
        if( hexVal.length() == 1 )
          hexVal = "0" + hexVal;
        outputArea.append( "Main Memory[$" + subTok1.toUpperCase() + "] set to $" + hexVal.toUpperCase() + "\n" );
      }
      // ------------------------------------------------------------------------------------------
      else
      {
        outputArea.append( "Unknown command: '" + keyTok + "'\n" );
      }

      outputArea.append( "\n" );
    }
  }

  /**
   * Initializes the output buffer.  This method clears the output buffer and
   * then displays a small header.
   */
  private void initialize()
  {
    outputArea.setText( "MonkeyNES OpCode Test Shell\n" );
    outputArea.append(  "Type 'help' for available commands\n\n" );

    outputArea.append( "Warning:  This test shell replaces $0000-$0002 with the requested\n" );
    outputArea.append( "          OpCode.  No effect of executing the OpCode is restored.\n\n" );
  }

  /**
   * Displays the OpCodeTestShellGui.
   */
  public void show()
  {
    // Position the window
    Dimension parentSize = parentFrame.getSize();
    Point     parentLoc  = parentFrame.getLocation();
    Dimension mySize     = new Dimension( 500, 400 );
    Point     myLoc      = new Point( parentLoc.x, parentLoc.y );

    // Fix the location in case the positioning put part of the window off the screen                                      
    if( myLoc.x < 0 ) myLoc.x = 0;                                      
    if( myLoc.y < 0 ) myLoc.y = 0;

    // Size and position the window
    mainDialog.setLocation( myLoc );
    mainDialog.setSize( mySize );

    initialize();

    // Show the whole mess...
    mainDialog.show();

    inputText.requestFocus();

    visible = true;
  }

}