package com.monkeycoder.monkeynes.gui;

import java.awt.*;
import javax.swing.*;

/**
 * This is a graphics component representing a boolean value as a light.
 * <br><br> 
 * Originally part of the MonkeyNES project.
 *
 * @author David Stephens
 */
public class StateLight extends JComponent
{
  private boolean   state            = false;
  private Color     trueColor        = null;
  private Color     trueBorderColor  = null;
  private Color     falseColor       = null;
  private Color     falseBorderColor = null;
  private Dimension totalSize        = new Dimension( 10, 16 );
  private Dimension lightSize        = new Dimension( 10, 10 );
  
  /** 
   * Creates a new StateLight using the specified colors for true and false.
   * @param inTrueColor The color to use when the light is representing true.
   * @param inFalseColor The color to use when the light is representing false.
   */
  public StateLight( Color inTrueColor, Color inFalseColor )
  {
    state = false;
    initColors( inTrueColor, inTrueColor, inFalseColor, inFalseColor );
  }

  /** 
   * Creates a new StateLight using the specified colors for true and false, and in the specified state.
   * @param inState The state this light is in upon creation.
   * @param inTrueColor The color to use when the light is representing true.
   * @param inFalseColor The color to use when the light is representing false.
   */  
  public StateLight( boolean inState, Color inTrueColor, Color inFalseColor )
  {
    state = inState;
    initColors( inTrueColor, inTrueColor, inFalseColor, inFalseColor );
  }

  /** 
   * Creates a new StateLight using the specified colors and border colors for true and false.
   * @param inTrueColor The color to use when the light is representing true.
   * @param inTrueBorderColor The border color to use when the light is representing true.
   * @param inFalseColor The color to use when the light is representing false.
   * @param inFalseBorderColor The border color to use when the light is representing false.
   */  
  public StateLight( Color inTrueColor, Color inTrueBorderColor, Color inFalseColor, Color inFalseBorderColor )
  {
    state = false;
    initColors( inTrueColor, inTrueBorderColor, inFalseColor, inFalseBorderColor );
  }

  /** 
   * Creates a new StateLight using the specified colors and border colors for true and false, and in the specified state.
   * @param inState The state this light is in upon creation.
   * @param inTrueColor The color to use when the light is representing true.
   * @param inTrueBorderColor The border color to use when the light is representing true.
   * @param inFalseColor The color to use when the light is representing false.
   * @param inFalseBorderColor The border color to use when the light is representing false.
   */  
  public StateLight( boolean inState, Color inTrueColor, Color inTrueBorderColor, Color inFalseColor, Color inFalseBorderColor )
  {
    state = inState;
    initColors( inTrueColor, inTrueBorderColor, inFalseColor, inFalseBorderColor );
  }

  /** 
   * Sets up the colors that will be used for this light.
   * @param inTrueColor The color to use when the light is representing true.
   * @param inTrueBorderColor The border color to use when the light is representing true.
   * @param inFalseColor The color to use when the light is representing false.
   * @param inFalseBorderColor The border color to use when the light is representing false.
   */  
  private void initColors( Color inTrueColor, Color inTrueBorderColor, Color inFalseColor, Color inFalseBorderColor )
  {
    trueColor        = inTrueColor;
    trueBorderColor  = inTrueBorderColor;
    falseColor       = inFalseColor;
    falseBorderColor = inFalseBorderColor;
    
    this.setSize( totalSize );
    this.setMinimumSize( totalSize );
    this.setMaximumSize( totalSize );
    this.setPreferredSize( totalSize );
  }

  /** 
   * Gets the current state of the light.
   * @return The current state of the light.
   */
  public boolean getState()
  {
    return state;
  }
  
  /** 
   * Sets the current state of the light.
   * @param inState The current state of the light.
   */
  public void setState( boolean inState )
  {
    state = inState;
    repaint();
  }

  /** 
   * Returns the size this light prefers to be the parent drawing component.
   * @return The size this light prefers to be.
   */  
  public Dimension getPreferredSize()
  {
    return totalSize;
  }

  /**
   * Paints the surface of this component based on the current state and colors.
   * @param g The graphics component that will be used for drawing this component.
   */   
  public void paint( Graphics g )
  {
    int top  = (int)(((double)totalSize.height / 2.0) - ((double)lightSize.height / 2.0));
    int left = (int)(((double)totalSize.width / 2.0) - ((double)lightSize.width / 2.0));
    
    if( state )
    {
      g.setColor( trueColor );
      g.fillOval( left, top, lightSize.width, lightSize.height );
      g.setColor( trueBorderColor );
      g.drawOval( left, top, lightSize.width-1, lightSize.height-1 );
    }
    else
    {
      g.setColor( falseColor );
      g.fillOval( left, top, lightSize.width, lightSize.height );
      g.setColor( falseBorderColor );
      g.drawOval( left, top, lightSize.width-1, lightSize.height-1 );
    }  
  }
}