package com.monkeycoder.monkeynes.hardware;

import java.util.*;

/**
 * This class acts like the Game Genie (&trade; of Galoob) game enhancement device.
 * Please note that MemoryFilters apply ONLY to single byte  reads and writes or 
 * byte range reads and writes.  No filtering is done for "word" reads or writes.
 * <br><br>
 * Originally part of the MonkeyNES project.
 *
 * @author David Stephens
 */
public class GameGenie implements MemoryFilter
{
  private ArrayList     activeCodes = new ArrayList();

  /**
   * Creates a new GameGenie.
   */
  public GameGenie()
  {
    // Do nothing...
  }
  
  /** 
   * Adds a code to the list of active codes.
   * @param inCode The code to add to the list of active codes.
   */
  public void addCode( GameGenieCode inCode )
  {
    activeCodes.add( inCode );
  }
  
  /** 
   * Removes a code from the list of active codes.
   * @param inIndex The index of the code to remove.
   */
  public void removeCode( int inIndex )
  {
    activeCodes.remove( inIndex );
  }

  /** 
   * Removes a code from the list of active codes.  The remove method that
   * uses index directly is recommended.
   * @param inCode The exact code object to remove.
   */
  public void removeCode( GameGenieCode inCode )
  {
    activeCodes.remove( activeCodes.indexOf(inCode) );
  }

  /** 
   * This function will mutate values read from memory according to 
   * the active GameGenieCode rules.
   * @param inAddress The address from which the value was read.
   * @param inValue A value already read from memory.
   * @return The filtered value at that address.
   */
  public int getFilteredReadValue( int inAddress, int inValue )
  {
    GameGenieCode tempCode = null;
    
    // Scan active codes to see if this filter applies
    for( int i = 0; i < activeCodes.size(); i++ )
    {
      tempCode = (GameGenieCode)activeCodes.get( i );
      
      if( inAddress == tempCode.getAddress() )
      {
        if( tempCode.usesTest() )
        {
          if( inValue == tempCode.getTest() )
            return tempCode.getValue();
          else
            return inValue;
        }
        else
          return tempCode.getValue();
      }
    }
    
    return inValue;
  }

  /** 
   * This function is not used by the GameGenie so it will always
   * return the value passed in with no further processing.
   * @param inAddress The address of the requested value.
   * @param inValue The value that was going to be written.
   * @return The filtered value to write at the indicated address.
   */
  public int getFilteredWriteValue( int inAddress, int inValue )
  {
    return inValue;
  }
}