package com.monkeycoder.monkeynes.hardware;

/**
 * This class holds a translated code, and provides static methods for
 * translating codes.
 * <br><br>
 * Originally part of the MonkeyNES project.
 *
 * @author David Stephens
 */
public class GameGenieCode
{
  /** 
   * A lookup array of letters used by the Original Game Genie.  The letters
   * are indexed by their values.
   */
  public static final String letters[] = { "A", "P", "Z", "L",
                                           "G", "I", "T", "Y",
                                           "E", "O", "X", "U",
                                           "K", "S", "V", "N" };

  private int     address  = 0;       // address for this change
  private int     value    = 0;       // new value to supply
  private int     test     = 0;       // old value must match test value for new value to be provided
  private boolean usesTest = false;   // whether or not the test value is used by this code

  /**
   * Creates a new GameGenieCode with the supplied address and value.  This is
   * the equivelent of a 6 character code.  A GameGenieCode created with this
   * function will report false for usesTest() and will unconditionally return
   * the inValue.
   * @param inAddress The 16-bit address that may be changed by this code.
   * @param inValue The value that will be provided instead of the memory value.
   * @see com.monkeycoder.monkeynes.hardware.GameGenieCode#usesTest()
   */
  public GameGenieCode( int inAddress, int inValue )
  {
    address  = inAddress;
    value    = inValue;
    usesTest = false;
  }

  /**
   * Creates a new GameGenieCode with the supplied address, value, and test value.
   * This is the equivelent of an 8 character code.  A GameGenieCode created with
   * this function will report true for usesTest().
   * @param inAddress The 16-bit address that may be changed by this code.
   * @param inValue The value that will be provided instead of the memory value.
   * @param inTest The value that the original memory value is compared against.
   *               If the value matches inValue is the response, otherwise the
   *               value in memory is returned unaltered.
   * @see com.monkeycoder.monkeynes.hardware.GameGenieCode#usesTest()
   */
  public GameGenieCode( int inAddress, int inValue, int inTest )
  {
    address  = inAddress;
    value    = inValue;
    test     = inTest;
    usesTest = true;
  }

  /**
   * Returns the 16-bit address that should be affected by this code.
   * @return The 16-bit address that should be affected by this code.
   */
  public int getAddress()
  {
    return address;
  }

  /**
   * Returns the value that should be returned instead of the value in memory.
   * @return The value that should be returned instead of the value in memory.
   */
  public int getValue()
  {
    return value;
  }

  /**
   * Returns the value that must be in memory at the specified address for the
   * altered value to be returned.
   * @return The test value.
   */
  public int getTest()
  {
    return test;
  }

  /**
   * Returns whether or not this code uses the test value.
   * @return Whether or not this code uses the test value.
   */
  public boolean usesTest()
  {
    return usesTest;
  }

  /**
   * Returns this GameGenieCode object as a code in the original form
   * used by the Game Genie.
   * @return An original Game Genie code.
   */
  public String getCode()
  {
    return GameGenieCode.translate( this );
  }

  /**
   * Creates a String with a reverse engineered code and the data values from this
   * GameGenieCode.
   * @return A String containing the resulting data.
   */  
  public String toString()
  {
    StringBuffer sb = new StringBuffer();
    
    sb.append( "Code:   " );
    sb.append( getCode() );
    sb.append( "\nAddr:   " );
    sb.append( Integer.toString( address, 16 ).toUpperCase() );
    sb.append( "\nValue:  " );
    sb.append( Integer.toString( value, 16 ).toUpperCase() );
    sb.append( "\nTest:   " );
    sb.append( Integer.toString( test, 16 ).toUpperCase() );   
    sb.append( "\n" );
    
    return sb.toString();
  }
  
  // ---------------------------------------------------------

  /**
   * Expects a code in the original form used by the Game Genie and will
   * translate the information into a GameGenieCode object.
   * @param inCode An original Game Genie code.
   * @return A GameGenieCode object with the translated information.
   */
  public static GameGenieCode translate( String inCode )
  {
    int[] nibs = getHex( inCode );
    int   tempAddress = 0;
    int   tempValue   = 0;
    int   tempTest    = 0;

    // address is the same for both versions
    tempAddress = (0x8 << 12)          | ((nibs[3] & 7) << 12) |
                  ((nibs[4] & 8) << 8) | ((nibs[5] & 7) << 8)  |
                  ((nibs[1] & 8) << 4) | ((nibs[2] & 7) << 4)  |
                  ( nibs[3] & 8)       |  (nibs[4] & 7);

    if( nibs.length == 6 )
    {
      tempValue = ((nibs[0] & 8) << 4) | ((nibs[1] & 7) << 4) |
                   (nibs[5] & 8)       |  (nibs[0] & 7);

      return new GameGenieCode( tempAddress, tempValue );
    }
    else if( nibs.length == 8 )
    {
      tempValue = ((nibs[0] & 8) << 4) | ((nibs[1] & 7) << 4) |
                   (nibs[7] & 8)       |  (nibs[0] & 7);
      
      tempTest  = ((nibs[6] & 8) << 4) | ((nibs[7] & 7) << 4) |
                   (nibs[5] & 8)       |  (nibs[6] & 7);                   
      
      return new GameGenieCode( tempAddress, tempValue, tempTest );             
    }

    return null;
  }

  /**
   * Translates any GameGenieCode object into a code in the original form
   * used by the Game Genie.
   * @param inCode A GameGenieCode object.
   * @return An original Game Genie code.
   */
  public static String translate( GameGenieCode inCode )
  {
    StringBuffer sb = new StringBuffer();
    int[] nibs = new int[8];
    
    // Assemble the nibbles
    if( inCode.usesTest() == false )
    {
      nibs[0] = ((inCode.getValue()   & 0x80)   >> 4) |  (inCode.getValue()   & 0x07);
      nibs[1] = ((inCode.getAddress() & 0x0080) >> 4) | ((inCode.getValue()   & 0x70)   >> 4);
      nibs[2] = (0x8)                                 | ((inCode.getAddress() & 0x0070) >> 4);
      nibs[3] =  (inCode.getAddress() & 0x0008)       | ((inCode.getAddress() & 0x7000) >> 12);
      nibs[4] = ((inCode.getAddress() & 0x0800) >> 8) |  (inCode.getAddress() & 0x0007);
      nibs[5] =  (inCode.getValue()   & 0x08)         | ((inCode.getAddress() & 0x0700) >> 8);
      
      sb.append( letters[nibs[0]] );
      sb.append( letters[nibs[1]] );
      sb.append( letters[nibs[2]] );
      sb.append( letters[nibs[3]] );
      sb.append( letters[nibs[4]] );
      sb.append( letters[nibs[5]] );
    }
    else
    {
      nibs[0] = ((inCode.getValue()   & 0x80)   >> 4) |  (inCode.getValue()   & 0x07);
      nibs[1] = ((inCode.getAddress() & 0x0080) >> 4) | ((inCode.getValue()   & 0x70)     >> 4);
      nibs[2] = (0x8)                                 | ((inCode.getAddress() & 0x0070)   >> 4);;
      nibs[3] =  (inCode.getAddress() & 0x0008)       | ((inCode.getAddress() & 0x7000)   >> 12);
      nibs[4] = ((inCode.getAddress() & 0x0800) >> 8) |  (inCode.getAddress() & 0x0007);
      nibs[5] =  (inCode.getTest()    & 0x08)         | ((inCode.getAddress() & 0x0700)   >> 8);
      nibs[6] = ((inCode.getTest()    & 0x80)   >> 4) |  (inCode.getTest()    & 0x07);
      nibs[7] =  (inCode.getValue()   & 0x08)         | ((inCode.getTest()    & 0x70)     >> 4);
                                              
      sb.append( letters[nibs[0]] );
      sb.append( letters[nibs[1]] );
      sb.append( letters[nibs[2]] );
      sb.append( letters[nibs[3]] );
      sb.append( letters[nibs[4]] );
      sb.append( letters[nibs[5]] );
      sb.append( letters[nibs[6]] );
      sb.append( letters[nibs[7]] );      
    }
    
    return sb.toString();
  }

  /**
   * Creates an array of integers representing the nibbles of a code in the
   * original form used by the Game Genie.
   * @param inCode An original Game Genie code.
   * @return An array of nibbles each as the low-order 4 bits in a separate integer.
   */
  private static int[] getHex( String inCode )
  {
    int[] nibbles = new int[inCode.length()];

    for( int i = 0; i < inCode.length(); i++ )
    {
      nibbles[i] = getLetterValue( inCode.substring( i, i+1 ).toUpperCase() );
    }

    return nibbles;
  }
  
  /**
   * Looks up the nibble value for a given letter.
   * @param inLetter A letter from an original Game Genie code.
   * @return An integer containing a nibble as the low-order 4 bits
   * or -1 if there was no match.
   */
  private static int getLetterValue( String inLetter )
  {
    for( int i = 0; i < letters.length; i++ )
    {
      if( inLetter.substring(0,1).equalsIgnoreCase( letters[i] ) )
        return i;
    }

    return -1;
  }
}