package com.monkeycoder.monkeynes.hardware;

import com.monkeycoder.monkeynes.base.*;

/**
 * This class represents the INES format ROM header.  It provides access to all 
 * the data available in the header.
 * <br><br> 
 * Originally part of the MonkeyNES project.
 *
 * @author David Stephens
 */
public class InesRomHeader
{
  private byte[] romHead = null;
  
  private static final int RH_PRG_ROM_COUNT     = 0x00000004;
  private static final int RH_CHR_ROM_COUNT     = 0x00000005;
  private static final int RH_ROM_CONTROL_1     = 0x00000006;
  private static final int RH_ROM_CONTROL_2     = 0x00000007;
  
  
  public static final int PRG_PAGE_SIZE         = 0x00004000;  // 16k
  public static final int CHR_PAGE_SIZE         = 0x00002000;  // 8k

  /**
   * Creates an INES Header using the bytes passed in.  These bytes
   * are probably loaded from a file, but can be loaded from anywhere.
   * @param inRomHeader The raw INES header data.
   */
  public InesRomHeader( byte[] inRomHeader )
  {
    setRomHeader( inRomHeader );
  }

  /**
   * Sets the raw data of an INES Header using the bytes passed in.  
   * These bytes are probably loaded from a file, but can be loaded from 
   * anywhere.
   * @param inRomHeader The raw INES header data.
   */  
  public void setRomHeader( byte[] inRomHeader )
  {
    romHead = inRomHeader;
  }
  
  /**
   * Gets the number of 16k pages of PRG-ROM data that are available.
   * @return The number of PRG-ROM pages available.
   */    
  public int getPrgRomPageCount()
  {
    return (int)romHead[RH_PRG_ROM_COUNT];
  } 
  
  /**
   * Gets the total size of the available PRG-ROM.
   * @return The size in bytes of the PRG-ROM.
   */ 
  public int getPrgRomSize()
  {
    return (int)romHead[RH_PRG_ROM_COUNT] * PRG_PAGE_SIZE;
  }   

  /**
   * Gets the number of 8k pages of CHR-ROM data that are available.
   * @return The number of CHR-ROM pages available.
   */ 
  public int getChrRomPageCount()
  {
    return (int)romHead[RH_CHR_ROM_COUNT];
  }

  /**
   * Gets the total size of the available CHR-ROM.
   * @return The size in bytes of the CHR-ROM.
   */   
  public int getChrRomSize()
  {
    return (int)romHead[RH_CHR_ROM_COUNT] * CHR_PAGE_SIZE;
  }  

  /**
   * Gets the number of the mapper in use by this cartrige.
   * @return The mapper number.
   */   
  public int getMapperNumber()
  {
    // shift the correct nibbles into place
    return (((int)romHead[RH_ROM_CONTROL_2]) & MathHelper.nibbleMask[1]) | 
           ((((int)romHead[RH_ROM_CONTROL_1]) & MathHelper.nibbleMask[1]) >> 4 );
  }
  
  /**
   * Gets whether or not there is 4-Screen VRam.
   * @return Whether or not there is 4-Screen VRam.
   */  
  public boolean isFourScreenVram()
  {
    return ( romHead[RH_ROM_CONTROL_1] & MathHelper.bitMask[3] ) == MathHelper.bitMask[3];
  }

  /**
   * Gets whether or not there is a trainer present.
   * @return Whether or not there is a trainer present.
   */  
  public boolean isTrainerPresent()
  {
    return ( romHead[RH_ROM_CONTROL_1] & MathHelper.bitMask[2] ) == MathHelper.bitMask[2];
  }

  /**
   * Gets whether or not SRam is enabled.
   * @return Whether or not SRam is enabled.
   */  
  public boolean isSramEnabled()
  {
    return ( romHead[RH_ROM_CONTROL_1] & MathHelper.bitMask[1] ) == MathHelper.bitMask[1];
  }    
  
  /**
   * Gets whether or not there is a Vertical Mirror.
   * @return Whether or not there is a Vertical Mirror.
   */
  public boolean isVertMirror()
  {
    return ( romHead[RH_ROM_CONTROL_1] & MathHelper.bitMask[0] ) == MathHelper.bitMask[0];
  }  
  
  /**
   * Gets whether or not there is a Horizontal Mirror.
   * @return Whether or not there is a Horizontal Mirror.
   */
  public boolean isHorzMirror()
  {
    return ( romHead[RH_ROM_CONTROL_1] & MathHelper.bitMask[0] ) == 0;
  }   
  
}