package com.monkeycoder.monkeynes.hardware;

import com.monkeycoder.monkeynes.base.*;

/**
 * This is the brain of the NES.  It contains registers, does the work of the OpCodes, 
 * and keeps track of machine cycles.
 * <br><br> 
 * Originally part of the MonkeyNES project.
 *
 * @author David Stephens
 */
public class NesCpu
{
  // Registers
  private CpuRegister16 pc  = new CpuRegister16();     // Program Counter (16-bit)
  private CpuRegister   sp  = new CpuRegister();       // Stack Pointer
  private CpuRegister   acc = new CpuRegister();       // Accumulator
  private CpuRegister   ix  = new CpuRegister();       // X-Index
  private CpuRegister   iy  = new CpuRegister();       // Y-Index
  private CpuRegister   ps  = new CpuRegister();       // Processor Status

  public static int STACK_PAGE   = 0x00000100;    // Page 1
  public static int NMI_VECTOR   = 0x0000FFFA;
  public static int RESET_VECTOR = 0x0000FFFC;
  public static int IRQ_VECTOR   = 0x0000FFFE;

  // Memory Banks
  private NesMainMemory  mainMem = null;
  private NesVideoMemory vidMem  = null;

  private int            needCycles  = -1;   // -1 means must calculate
  private int            availCycles = 0;

  /**
   * Creates a new NesCpu.
   * @param inMainMem Main memory bank accessed by this CPU.
   * @param inVidMem Video memory bank accessed by this CPU.
   */
  public NesCpu( NesMainMemory inMainMem, NesVideoMemory inVidMem )
  {
    mainMem = inMainMem;
    vidMem  = inVidMem;
  }

  /**
   * Called to tell the CPU that power is now on.  This sets up register values and 
   * prepares the system for operation.
   */
  public void powerOn()
  {
    pc.setValue(  (mainMem.readNoMessage( RESET_VECTOR + 1 ) << 8) | mainMem.readNoMessage( RESET_VECTOR ) );
    sp.setValue(  0x000000FF );
    acc.setValue( 0x00000000 );
    ix.setValue(  0x00000000 );
    iy.setValue(  0x00000000 );
    ps.setValue(  0x00000020 );
  }

  /**
   * Called to tell the CPU that power is now off.  This clears register values and 
   * shuts down the system.
   */
  public void powerOff()
  {
    pc.setValue(  (mainMem.readNoMessage( RESET_VECTOR + 1 ) << 8) | mainMem.readNoMessage( RESET_VECTOR ) );
    sp.setValue(  0x000000FF );
    acc.setValue( 0x00000000 );
    ix.setValue(  0x00000000 );
    iy.setValue(  0x00000000 );
    ps.setValue(  0x00000020 );
  }

  /**
   * Called to tell the CPU to do a cycle.  Unlike most other cycle dependent hardware in
   * the system, the CPU requires different numbers of cycles depending on what it's
   * currently working on.  This is all tracked internally.
   */
  public void doCycle()
  {
    availCycles++;

    if( availCycles >= needCycles )
      executeNextInstruction();
  }

  /**
   * Called at the completion of an OpCode, this method determines how many 
   * cycles will be needed for the next operation.  It takes into account extra 
   * cycles for crossing page boundries and extra cycles for taking a branch as 
   * applicible.
   * @return The number of cycles that will be required for the next operation.
   */
  private int getNextInstructionCycles()
  {
    int curOp = mainMem.readNoMessage( pc.getValue() );
    int cycles = 0;

    switch( curOp )
    {
      // Instructions with nothing odd about them for timing -------------
      case 0x00:  // BRK (Implied)
      case 0x01:  // ORA (Indirect,X)
      case 0x05:  // ORA (Zero Page)
      case 0x06:  // ASL (Zero Page)
      case 0x08:  // PHP (Implied)
      case 0x09:  // ORA (Immediate)
      case 0x0A:  // ASL (Accumulator)
      case 0x0D:  // ORA (Absolute)
      case 0x0E:  // ASL (Absolute)
      case 0x15:  // ORA (Zero Page,X)
      case 0x16:  // ASL (Zero Page,X)
      case 0x18:  // CLC (Implied)
      case 0x1E:  // ASL (Absolute,X)
      case 0x20:  // JSR (Absolute)
      case 0x21:  // AND (Indirect,X)
      case 0x24:  // BIT (Zero Page)
      case 0x25:  // AND (Zero Page)
      case 0x26:  // ROL (Zero Page)
      case 0x28:  // PLP (Implied)
      case 0x29:  // AND (Immediate)
      case 0x2A:  // ROL (Accumulator)
      case 0x2C:  // BIT (Absolute)
      case 0x2D:  // AND (Absolute)
      case 0x2E:  // ROL (Absolute)
      case 0x35:  // AND (Zero Page,X)
      case 0x36:  // ROL (Zero Page,X)
      case 0x38:  // SEC (Implied)
      case 0x3E:  // ROL (Absolute,X)
      case 0x40:  // RTI (Implied)
      case 0x41:  // EOR (Indirect,X)
      case 0x45:  // EOR (Zero Page)
      case 0x46:  // LSR (Zero Page)
      case 0x48:  // PHA (Implied)
      case 0x49:  // EOR (Immediate)
      case 0x4A:  // LSR (Accumulator)
      case 0x4C:  // JMP (Absolute)
      case 0x4D:  // EOR (Absolute)
      case 0x4E:  // LSR (Absolute)
      case 0x55:  // EOR (Zero Page,X)
      case 0x56:  // LSR (Zero Page,X)
      case 0x58:  // CLI (Implied)
      case 0x5E:  // LSR (Absolute,X)
      case 0x60:  // RTS (Implied)
      case 0x61:  // ADC (Indirect,X)
      case 0x65:  // ADC (Zero Page)
      case 0x66:  // ROR (Zero Page)
      case 0x68:  // PLA (Implied)
      case 0x69:  // ADC (Immediate)
      case 0x6A:  // ROR (Accumulator)
      case 0x6C:  // JMP (Indirect)
      case 0x6D:  // ADC (Absolute)
      case 0x6E:  // ROR (Absolute)
      case 0x75:  // ADC (Zero Page,X)
      case 0x76:  // ROR (Zero Page,X)
      case 0x78:  // SEI (Implied)
      case 0x7C:  // JMP (Absolute,X)
      case 0x7E:  // ROR (Absolute,X)
      case 0x81:  // STA (Indirect,X)
      case 0x84:  // STY (Zero Page)
      case 0x85:  // STA (Zero Page)
      case 0x86:  // STX (Zero Page)
      case 0x88:  // DEY (Implied)
      case 0x8A:  // TXA (Implied)
      case 0x8C:  // STY (Absolute)
      case 0x8D:  // STA (Absolute)
      case 0x8E:  // STX (Absolute)
      case 0x91:  // STA (Indirect,Y)
      case 0x94:  // STY (Zero Page,X)
      case 0x95:  // STA (Zero Page,X)
      case 0x96:  // STX (Zero Page,Y)
      case 0x98:  // TYA (Implied)
      case 0x99:  // STA (Absolute,Y)
      case 0x9A:  // TXS (Implied)
      case 0x9D:  // STA (Absolute,X)
      case 0xA0:  // LDY (Immediate)
      case 0xA1:  // LDA (Indirect,X)
      case 0xA2:  // LDX (Immediate)
      case 0xA4:  // LDY (Zero Page)
      case 0xA5:  // LDA (Zero Page)
      case 0xA6:  // LDX (Zero Page)
      case 0xA8:  // TAY (Implied)
      case 0xA9:  // LDA (Immediate)
      case 0xAA:  // TAX (Implied)
      case 0xAC:  // LDY (Absolute)
      case 0xAD:  // LDA (Absolute)
      case 0xAE:  // LDX (Absolute)
      case 0xB4:  // LDY (Zero Page,X)
      case 0xB5:  // LDA (Zero Page,X)
      case 0xB6:  // LDX (Zero Page,Y)
      case 0xB8:  // CLV (Implied)
      case 0xBA:  // TSX (Implied)
      case 0xC0:  // CPY (Immediate)
      case 0xC1:  // CMP (Indirect,X)
      case 0xC4:  // CPY (Zero Page)
      case 0xC5:  // CMP (Zero Page)
      case 0xC6:  // DEC (Zero Page)
      case 0xC8:  // INY (Implied)
      case 0xC9:  // CMP (Immediate)
      case 0xCA:  // DEX (Implied)
      case 0xCC:  // CPY (Absolute)
      case 0xCD:  // CMP (Absolute)
      case 0xCE:  // DEC (Absolute)
      case 0xD5:  // CMP (Zero Page,X)
      case 0xD6:  // DEC (Zero Page,X)
      case 0xD8:  // CLD (Implied)
      case 0xDE:  // DEC (Absolute,X)
      case 0xE0:  // CPX (Immediate)
      case 0xE1:  // SBC (Indirect,X)
      case 0xE4:  // CPX (Zero Page)
      case 0xE5:  // SBC (Zero Page)
      case 0xE6:  // INC (Zero Page)
      case 0xE8:  // INX (Implied)
      case 0xE9:  // SBC (Immediate)
      case 0xEA:  // NOP (Implied)
      case 0xEC:  // CPX (Absolute)
      case 0xED:  // SBC (Absolute)
      case 0xEE:  // INC (Absolute)
      case 0xF5:  // SBC (Zero Page,X)
      case 0xF6:  // INC (Zero Page,X)
      case 0xF8:  // SED (Implied)
      case 0xFE:  // INC (Absolute,X)
        cycles = NesOpCodeTable.cycles[curOp];
        break;

      // Instructions with extra cycles for page boundry crossing (1 byte params) --------
      case 0x11:  // ORA (Indirect,Y)
      case 0x31:  // AND (Indirect,Y)
      case 0x51:  // EOR (Indirect,Y)
      case 0x71:  // ADC (Indirect,Y)
      case 0xB1:  // LDA (Indirect,Y)
      case 0xD1:  // CMP (Indirect,Y)
      case 0xF1:  // SBC (Indirect,Y)      
        if( crossesPageBoundry( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) )
          cycles = NesOpCodeTable.cycles[curOp] + NesOpCodeTable.pb_cycles[curOp];
        else
          cycles = NesOpCodeTable.cycles[curOp];
        break;  
      
      // Instructions with extra cycles for page boundry crossing (2 byte params) --------                        
      case 0x19:  // ORA (Absolute,Y)
      case 0x1D:  // ORA (Absolute,X)
      case 0x39:  // AND (Absolute,Y)
      case 0x3D:  // AND (Absolute,X)
      case 0x59:  // EOR (Absolute,Y)
      case 0x5D:  // EOR (Absolute,X)
      case 0x79:  // ADC (Absolute,Y)
      case 0x7D:  // ADC (Absolute,X)
      case 0xB9:  // LDA (Absolute,Y)
      case 0xBC:  // LDY (Absolute,X)
      case 0xBD:  // LDA (Absolute,X)
      case 0xBE:  // LDX (Absolute,Y)
      case 0xD9:  // CMP (Absolute,Y)
      case 0xDD:  // CMP (Absolute,X)
      case 0xF9:  // SBC (Absolute,Y)
      case 0xFD:  // SBC (Absolute,X)
        if( crossesPageBoundry( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) )
          cycles = NesOpCodeTable.cycles[curOp] + NesOpCodeTable.pb_cycles[curOp];
        else
          cycles = NesOpCodeTable.cycles[curOp];
        break;  

      // Branch Instructions ---------------------------------------------
      case 0x10:  // BPL (Relative)
      case 0x30:  // BMI (Relative)
      case 0x50:  // BVC (Relative)
      case 0x70:  // BVS (Relative)
      case 0x90:  // BCC (Relative)
      case 0xB0:  // BCS (Relative)
      case 0xD0:  // BNE (Relative)
      case 0xF0:  // BEQ (Relative)
        cycles = NesOpCodeTable.cycles[curOp];
        if( crossesPageBoundry( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) )
          cycles += NesOpCodeTable.pb_cycles[curOp];
        if( willBranch( curOp ) )
          cycles += NesOpCodeTable.bt_cycles[curOp];
        break;  

      // Unknown Op Codes ------------------------------------------------
      case 0x02:  //
      case 0x03:  //
      case 0x04:  //
      case 0x07:  //
      case 0x0B:  //
      case 0x0C:  //
      case 0x0F:  //
      case 0x12:  //
      case 0x13:  //
      case 0x14:  //
      case 0x17:  //
      case 0x1A:  //
      case 0x1B:  //
      case 0x1C:  //
      case 0x1F:  //
      case 0x22:  //
      case 0x23:  //
      case 0x27:  //
      case 0x2B:  //
      case 0x2F:  //
      case 0x32:  //
      case 0x33:  //
      case 0x34:  //
      case 0x37:  //
      case 0x3A:  //
      case 0x3B:  //
      case 0x3C:  //
      case 0x3F:  //
      case 0x42:  //
      case 0x43:  //
      case 0x44:  //
      case 0x47:  //
      case 0x4B:  //
      case 0x4F:  //
      case 0x52:  //
      case 0x53:  //
      case 0x54:  //
      case 0x57:  //
      case 0x5A:  //
      case 0x5B:  //
      case 0x5C:  //
      case 0x5F:  //
      case 0x62:  //
      case 0x63:  //
      case 0x64:  //
      case 0x67:  //
      case 0x6B:  //
      case 0x6F:  //
      case 0x72:  //
      case 0x73:  //
      case 0x74:  //
      case 0x77:  //
      case 0x7A:  //
      case 0x7B:  //
      case 0x7F:  //
      case 0x80:  //
      case 0x82:  //
      case 0x83:  //
      case 0x87:  //
      case 0x89:  //
      case 0x8B:  //
      case 0x8F:  //
      case 0x92:  //
      case 0x93:  //
      case 0x97:  //
      case 0x9B:  //
      case 0x9C:  //
      case 0x9E:  //
      case 0x9F:  //
      case 0xA3:  //
      case 0xA7:  //
      case 0xAB:  //
      case 0xAF:  //
      case 0xB2:  //
      case 0xB3:  //
      case 0xB7:  //
      case 0xBB:  //
      case 0xBF:  //
      case 0xC2:  //
      case 0xC3:  //
      case 0xC7:  //
      case 0xCB:  //
      case 0xCF:  //
      case 0xD2:  //
      case 0xD3:  //
      case 0xD4:  //
      case 0xD7:  //
      case 0xDA:  //
      case 0xDB:  //
      case 0xDC:  //
      case 0xDF:  //
      case 0xE2:  //
      case 0xE3:  //
      case 0xE7:  //
      case 0xEB:  //
      case 0xEF:  //
      case 0xF2:  //
      case 0xF3:  //
      case 0xF4:  //
      case 0xF7:  //
      case 0xFA:  //
      case 0xFB:  //
      case 0xFC:  //
      case 0xFF:  //
      default:    // or anything else lost in the shuffle
        return 2;
    }
    
    return cycles;
  }

  /**
   * This is the work horse of the CPU.  This function handles the actual 
   * logic of all 152 OpCodes.
   */
  private void executeNextInstruction()
  {
    int curOp    = mainMem.readNoMessage( pc.getValue() );
    int addr     = 0;  // temp variable
    int tempVal  = 0;  // another temp variable
    int tempVal2 = 0;  // another temp variable
    int origVal  = 0;  // another temp variable

    switch( curOp )
    {
      case 0x00:    // BRK (Implied)
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        mainMem.writeNoMessage( STACK_PAGE + sp.getValue(), ((MathHelper.byteMask[1] & pc.getValue()) >> 8) );
        sp.decrement();
        mainMem.writeNoMessage( STACK_PAGE + sp.getValue(), (MathHelper.byteMask[0] & pc.getValue()) );
        sp.decrement();
        ps.setBitValue( CpuRegister.PS_BREAK_COMMAND, true );
        mainMem.writeNoMessage( STACK_PAGE + sp.getValue(), ps.getValue() );
        sp.decrement();
        ps.setBitValue( CpuRegister.PS_INTERRUPT_DISABLE, true );
        pc.setValue(  (mainMem.readNoMessage( IRQ_VECTOR + 1 ) << 8) | mainMem.readNoMessage( IRQ_VECTOR ) );
        break;
      case 0x01:    // ORA (Indirect,X)
        acc.orEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x02:    //
        DebugLog.log( 3, "op [02] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x03:    //
        DebugLog.log( 3, "op [03] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x04:    //
        DebugLog.log( 3, "op [04] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x05:    // ORA (Zero Page)
        acc.orEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x06:    // ASL (Zero Page)
        addr    = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal = mainMem.readNoMessage( addr );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG, (tempVal & MathHelper.bitMask[7]) );
        tempVal = (tempVal << 1) & MathHelper.byteMask[0];
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x07:    //
        // Absolutely no idea what this is, but there are like a billion of them in 
        // SMB1 and the debug output for it takes to long to do anything else.
        // DebugLog.log( 3, "op [07] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x08:    // PHP (Implied)
        mainMem.writeNoMessage( STACK_PAGE + sp.getValue(), ps.getValue() );
        sp.decrement();
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x09:    // ORA (Immediate)
        acc.orEquals( mainMem.readNoMessage( pc.getValue() + 1 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x0A:    // ASL (Accumulator)
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        acc.setValue( acc.getValue() << 1 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x0B:    //
        DebugLog.log( 3, "op [0B] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x0C:    //
        DebugLog.log( 3, "op [0C] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x0D:    // ORA (Absolute)
        acc.orEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x0E:    // ASL (Absolute)
        addr    = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        tempVal = mainMem.readNoMessage( addr );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG, (tempVal & MathHelper.bitMask[7]) );
        tempVal = (tempVal << 1) & MathHelper.byteMask[0];
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x0F:    //
        DebugLog.log( 3, "op [0F] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x10:    // BPL (Relative)
        if( willBranch( curOp ) )
          pc.plusEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + NesOpCodeTable.length[curOp] );
        else
          pc.plusEquals( NesOpCodeTable.length[curOp] );  
        break;
      case 0x11:    // ORA (Indirect,Y)
        acc.orEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x12:    //
        DebugLog.log( 3, "op [12] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x13:    //
        DebugLog.log( 3, "op [13] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x14:    //
        DebugLog.log( 3, "op [14] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x15:    // ORA (Zero Page,X)
        acc.orEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x16:    // ASL (Zero Page,X)
        addr    = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal = mainMem.readNoMessage( addr );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG, (tempVal & MathHelper.bitMask[7]) );
        tempVal = (tempVal << 1) & MathHelper.byteMask[0];
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x17:    //
        DebugLog.log( 3, "op [17] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x18:    // CLC (Implied)
        ps.setBitOff( CpuRegister.PS_CARRY_FLAG );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x19:    // ORA (Absolute,Y)
        acc.orEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x1A:    //
        DebugLog.log( 3, "op [1A] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x1B:    //
        DebugLog.log( 3, "op [1B] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x1C:    //
        DebugLog.log( 3, "op [1C] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x1D:    // ORA (Absolute,X)
        acc.orEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x1E:    // ASL (Absolute,X)
        addr    = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        tempVal = mainMem.readNoMessage( addr );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG, (tempVal & MathHelper.bitMask[7]) );
        tempVal = (tempVal << 1) & MathHelper.byteMask[0];
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x1F:    //
        DebugLog.log( 3, "op [1F] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x20:    // JSR (Absolute)
        pc.setValue( pc.getValue() + NesOpCodeTable.length[curOp] - 1 );
        mainMem.writeNoMessage( STACK_PAGE + sp.getValue(), ((MathHelper.byteMask[1] & pc.getValue()) >> 8) );
        sp.decrement();
        mainMem.writeNoMessage( STACK_PAGE + sp.getValue(), (MathHelper.byteMask[0] & pc.getValue()) );
        sp.decrement();
        pc.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        break;
      case 0x21:    // AND (Indirect,X)
        acc.andEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x22:    //
        DebugLog.log( 3, "op [22] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x23:    //
        DebugLog.log( 3, "op [23] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x24:    // BIT (Zero Page)
        tempVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG, ((acc.getValue() & tempVal) == 0) );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, ((tempVal & MathHelper.bitMask[7]) == MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, ((tempVal & MathHelper.bitMask[6]) == MathHelper.bitMask[6]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x25:    // AND (Zero Page)
        acc.andEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x26:    // ROL (Zero Page)
        addr     = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal  = mainMem.readNoMessage( addr );
        tempVal2 = ps.getBitValue( CpuRegister.PS_CARRY_FLAG );    // I'll need this later
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    (tempVal & MathHelper.bitMask[7]) );
        tempVal = (tempVal << 1) & MathHelper.byteMask[0];
        tempVal |= tempVal2;  // Put the saved carry flag into place
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x27:    //
        DebugLog.log( 3, "op [27] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x28:    // PLP (Implied)
        sp.increment();
        ps.setValue( mainMem.readNoMessage( STACK_PAGE + sp.getValue() ) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x29:    // AND (Immediate)
        acc.andEquals( mainMem.readNoMessage( pc.getValue() + 1 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x2A:    // ROL (Accumulator)
        tempVal2 = ps.getBitValue( CpuRegister.PS_CARRY_FLAG );    // I'll need this later
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        acc.setValue( acc.getValue() << 1 );
        acc.setBitValue( MathHelper.bitMask[0], tempVal2 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x2B:    //
        DebugLog.log( 3, "op [2B] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x2C:    // BIT (Absolute)
        tempVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG, ((acc.getValue() & tempVal) == 0) );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, ((tempVal & MathHelper.bitMask[7]) == MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, ((tempVal & MathHelper.bitMask[6]) == MathHelper.bitMask[6]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x2D:    // AND (Absolute)
        acc.andEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x2E:    // ROL (Absolute)
        addr     = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        tempVal  = mainMem.readNoMessage( addr );
        tempVal2 = ps.getBitValue( CpuRegister.PS_CARRY_FLAG );    // I'll need this later
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    (tempVal & MathHelper.bitMask[7]) );
        tempVal = (tempVal << 1) & MathHelper.byteMask[0];
        tempVal |= tempVal2;  // Put the saved carry flag into place
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x2F:    //
        DebugLog.log( 3, "op [2F] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x30:    // BMI (Relative)
        if( willBranch( curOp ) )
          pc.plusEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + NesOpCodeTable.length[curOp] );
        else
          pc.plusEquals( NesOpCodeTable.length[curOp] );  
        break;
      case 0x31:    // AND (Indirect,Y)
        acc.andEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x32:    //
        DebugLog.log( 3, "op [32] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x33:    //
        DebugLog.log( 3, "op [33] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x34:    //
        DebugLog.log( 3, "op [34] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x35:    // AND (Zero Page,X)
        acc.andEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x36:    // ROL (Zero Page,X)
        addr     = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal  = mainMem.readNoMessage( addr );
        tempVal2 = ps.getBitValue( CpuRegister.PS_CARRY_FLAG );    // I'll need this later
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    (tempVal & MathHelper.bitMask[7]) );
        tempVal = (tempVal << 1) & MathHelper.byteMask[0];
        tempVal |= tempVal2;  // Put the saved carry flag into place
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x37:    //
        DebugLog.log( 3, "op [37] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x38:    // SEC (Implied)
        ps.setBitOn( CpuRegister.PS_CARRY_FLAG );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x39:    // AND (Absolute,Y)
        acc.andEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x3A:    //
        DebugLog.log( 3, "op [3A] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x3B:    //
        DebugLog.log( 3, "op [3B] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x3C:    //
        DebugLog.log( 3, "op [3C] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x3D:    // AND (Absolute,X)
        acc.andEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x3E:    // ROL (Absolute,X)
        addr     = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        tempVal  = mainMem.readNoMessage( addr );
        tempVal2 = ps.getBitValue( CpuRegister.PS_CARRY_FLAG );    // I'll need this later
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    (tempVal & MathHelper.bitMask[7]) );
        tempVal = (tempVal << 1) & MathHelper.byteMask[0];
        tempVal |= tempVal2;  // Put the saved carry flag into place
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x3F:    //
        DebugLog.log( 3, "op [3F] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x40:    // RTI (Implied)
        sp.increment();
        ps.setValue( mainMem.readNoMessage( STACK_PAGE + sp.getValue() ) );
        sp.increment();
        addr = mainMem.readNoMessage( STACK_PAGE + sp.getValue() );
        sp.increment();
        addr |= (mainMem.readNoMessage( STACK_PAGE + sp.getValue() ) << 8);
        pc.setValue( addr );
        break;
      case 0x41:    // EOR (Indirect,X)
        acc.xorEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x42:    //
        DebugLog.log( 3, "op [42] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x43:    //
        DebugLog.log( 3, "op [43] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x44:    //
        DebugLog.log( 3, "op [44] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x45:    // EOR (Zero Page)
        acc.xorEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x46:    // LSR (Zero Page)
        addr    = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal = mainMem.readNoMessage( addr );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG, (tempVal & MathHelper.bitMask[0]) );
        tempVal = tempVal >> 1;
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x47:    //
        DebugLog.log( 3, "op [47] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x48:    // PHA (Implied)
        mainMem.writeNoMessage( STACK_PAGE + sp.getValue(), acc.getValue() );
        sp.decrement();
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x49:    // EOR (Immediate)
        acc.xorEquals( mainMem.readNoMessage( pc.getValue() + 1 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x4A:    // LSR (Accumulator)
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG, acc.getBitValue(MathHelper.bitMask[0]) );
        acc.setValue( acc.getValue() >> 1 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x4B:    //
        DebugLog.log( 3, "op [4B] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x4C:    // JMP (Absolute)
        pc.setValue( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        break;
      case 0x4D:    // EOR (Absolute)
        acc.xorEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x4E:    // LSR (Absolute)
        addr    = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        tempVal = mainMem.readNoMessage( addr );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG, (tempVal & MathHelper.bitMask[0]) );
        tempVal = tempVal >> 1;
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x4F:    //
        DebugLog.log( 3, "op [4F] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x50:    // BVC (Relative)
        if( willBranch( curOp ) )
          pc.plusEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + NesOpCodeTable.length[curOp] );
        else
          pc.plusEquals( NesOpCodeTable.length[curOp] );  
        break;
      case 0x51:    // EOR (Indirect,Y)
        acc.xorEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x52:    //
        DebugLog.log( 3, "op [52] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x53:    //
        DebugLog.log( 3, "op [53] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x54:    //
        DebugLog.log( 3, "op [54] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x55:    // EOR (Zero Page,X)
        acc.xorEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x56:    // LSR (Zero Page,X)
        addr    = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal = mainMem.readNoMessage( addr );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG, (tempVal & MathHelper.bitMask[0]) );
        tempVal = tempVal >> 1;
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x57:    //
        DebugLog.log( 3, "op [57] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x58:    // CLI (Implied)
        ps.setBitOff( CpuRegister.PS_INTERRUPT_DISABLE );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x59:    // EOR (Absolute,Y)
        acc.xorEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x5A:    //
        DebugLog.log( 3, "op [5A] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x5B:    //
        DebugLog.log( 3, "op [5B] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x5C:    //
        DebugLog.log( 3, "op [5C] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x5D:    // EOR (Absolute,X)
        acc.xorEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x5E:    // LSR (Absolute,X)
        addr    = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        tempVal = mainMem.readNoMessage( addr );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG, (tempVal & MathHelper.bitMask[0]) );
        tempVal = tempVal >> 1;
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x5F:    //
        DebugLog.log( 3, "op [5F] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x60:    // RTS (Implied)
        sp.increment();
        addr = mainMem.readNoMessage( STACK_PAGE + sp.getValue() );
        sp.increment();
        addr |= (mainMem.readNoMessage( STACK_PAGE + sp.getValue() ) << 8);
        pc.setValue( addr + 1 );
        break;
      case 0x61:    // ADC (Indirect,X)
        tempVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + acc.getValue() + ps.getBitValue( CpuRegister.PS_CARRY_FLAG );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    tempVal > 255 );
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x62:    //
        DebugLog.log( 3, "op [62] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x63:    //
        DebugLog.log( 3, "op [63] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x64:    //
        DebugLog.log( 3, "op [64] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x65:    // ADC (Zero Page)
        tempVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + acc.getValue() + ps.getBitValue( CpuRegister.PS_CARRY_FLAG );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    tempVal > 255 );
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x66:    // ROR (Zero Page)
        addr     = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal  = mainMem.readNoMessage( addr );
        tempVal2 = ps.getBitValue( CpuRegister.PS_CARRY_FLAG );    // I'll need this later
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    (tempVal & MathHelper.bitMask[0]) );
        tempVal = tempVal >> 1;
        tempVal |= (tempVal2 << 7);  // Put the saved carry flag into place
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x67:    //
        DebugLog.log( 3, "op [67] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x68:    // PLA (Implied)
        sp.increment();
        acc.setValue( mainMem.readNoMessage( STACK_PAGE + sp.getValue() ) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x69:    // ADC (Immediate)
        tempVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + acc.getValue() + ps.getBitValue( CpuRegister.PS_CARRY_FLAG );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    tempVal > 255 );
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x6A:    // ROR (Accumulator)
        tempVal2 = ps.getBitValue( CpuRegister.PS_CARRY_FLAG );    // I'll need this later
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG, acc.getBitValue(MathHelper.bitMask[0]) );
        acc.setValue( acc.getValue() >> 1 );
        acc.setBitValue( MathHelper.bitMask[7], tempVal2 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x6B:    //
        DebugLog.log( 3, "op [6B] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x6C:    // JMP (Indirect)
        pc.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        break;
      case 0x6D:    // ADC (Absolute)
        tempVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) + acc.getValue() + ps.getBitValue( CpuRegister.PS_CARRY_FLAG );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    tempVal > 255 );
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x6E:    // ROR (Absolute)
        addr     = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        tempVal  = mainMem.readNoMessage( addr );
        tempVal2 = ps.getBitValue( CpuRegister.PS_CARRY_FLAG );    // I'll need this later
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    (tempVal & MathHelper.bitMask[0]) );
        tempVal = tempVal >> 1;
        tempVal |= (tempVal2 << 7);  // Put the saved carry flag into place
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x6F:    //
        DebugLog.log( 3, "op [6F] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x70:    // BVS (Relative)
        if( willBranch( curOp ) )
          pc.plusEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + NesOpCodeTable.length[curOp] );
        else
          pc.plusEquals( NesOpCodeTable.length[curOp] );  
        break;
      case 0x71:    // ADC (Indirect,Y)
        tempVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + acc.getValue() + ps.getBitValue( CpuRegister.PS_CARRY_FLAG );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    tempVal > 255 );
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x72:    //
        DebugLog.log( 3, "op [72] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x73:    //
        DebugLog.log( 3, "op [73] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x74:    //
        DebugLog.log( 3, "op [74] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x75:    // ADC (Zero Page,X)
        tempVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + acc.getValue() + ps.getBitValue( CpuRegister.PS_CARRY_FLAG );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    tempVal > 255 );
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x76:    // ROR (Zero Page,X)
        addr     = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal  = mainMem.readNoMessage( addr );
        tempVal2 = ps.getBitValue( CpuRegister.PS_CARRY_FLAG );    // I'll need this later
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    (tempVal & MathHelper.bitMask[0]) );
        tempVal = tempVal >> 1;
        tempVal |= (tempVal2 << 7);  // Put the saved carry flag into place
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x77:    //
        DebugLog.log( 3, "op [77] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x78:    // SEI (Implied)
        ps.setBitOn( CpuRegister.PS_INTERRUPT_DISABLE );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x79:    // ADC (Absolute,Y)
        tempVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) + acc.getValue() + ps.getBitValue( CpuRegister.PS_CARRY_FLAG );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    tempVal > 255 );
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x7A:    //
        DebugLog.log( 3, "op [7A] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x7B:    //
        DebugLog.log( 3, "op [7B] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x7C:    // JMP (Absolute,X)
        pc.setValue( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        break;
      case 0x7D:    // ADC (Absolute,X)
        tempVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) + acc.getValue() + ps.getBitValue( CpuRegister.PS_CARRY_FLAG );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    tempVal > 255 );
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x7E:    // ROR (Absolute,X)
        addr     = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        tempVal  = mainMem.readNoMessage( addr );
        tempVal2 = ps.getBitValue( CpuRegister.PS_CARRY_FLAG );    // I'll need this later
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    (tempVal & MathHelper.bitMask[0]) );
        tempVal = tempVal >> 1;
        tempVal |= (tempVal2 << 7);  // Put the saved carry flag into place
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, (tempVal & MathHelper.bitMask[7]) );
        mainMem.writeNoMessage( addr, tempVal );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x7F:    //
        DebugLog.log( 3, "op [7F] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x80:    //
        DebugLog.log( 3, "op [80] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x81:    // STA (Indirect,X)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ), acc.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x82:    //
        DebugLog.log( 3, "op [82] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x83:    //
        DebugLog.log( 3, "op [83] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x84:    // STY (Zero Page)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ), iy.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x85:    // STA (Zero Page)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ), acc.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x86:    // STX (Zero Page)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ), ix.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x87:    //
        DebugLog.log( 3, "op [87] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x88:    // DEY (Implied)
        iy.decrement();
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     iy.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, iy.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x89:    //
        DebugLog.log( 3, "op [89] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x8A:    // TXA (Implied)
        acc.setValue( ix.getValue() );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x8B:    //
        DebugLog.log( 3, "op [8B] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x8C:    // STY (Absolute)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ), iy.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x8D:    // STA (Absolute)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ), acc.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x8E:    // STX (Absolute)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ), ix.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x8F:    //
        DebugLog.log( 3, "op [8F] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x90:    // BCC (Relative)
        if( willBranch( curOp ) )
          pc.plusEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + NesOpCodeTable.length[curOp] );
        else
          pc.plusEquals( NesOpCodeTable.length[curOp] );  
        break;
      case 0x91:    // STA (Indirect,Y)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ), acc.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x92:    //
        DebugLog.log( 3, "op [92] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x93:    //
        DebugLog.log( 3, "op [93] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x94:    // STY (Zero Page,X)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ), iy.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x95:    // STA (Zero Page,X)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ), acc.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x96:    // STX (Zero Page,Y)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ), ix.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x97:    //
        DebugLog.log( 3, "op [97] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x98:    // TYA (Implied)
        acc.setValue( iy.getValue() );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x99:    // STA (Absolute,Y)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ), acc.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x9A:    // TXS (Implied)
        sp.setValue( ix.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x9B:    //
        DebugLog.log( 3, "op [9B] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x9C:    //
        DebugLog.log( 3, "op [9C] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x9D:    // STA (Absolute,X)
        mainMem.writeNoMessage( getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ), acc.getValue() );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x9E:    //
        DebugLog.log( 3, "op [9E] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0x9F:    //
        DebugLog.log( 3, "op [9F] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xA0:    // LDY (Immediate)
        iy.setValue( mainMem.readNoMessage( pc.getValue() + 1 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     iy.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, iy.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xA1:    // LDA (Indirect,X)
        acc.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xA2:    // LDX (Immediate)
        ix.setValue( mainMem.readNoMessage( pc.getValue() + 1 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     ix.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, ix.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xA3:    //
        DebugLog.log( 3, "op [A3] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xA4:    // LDY (Zero Page)
        iy.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     iy.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, iy.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xA5:    // LDA (Zero Page)
        acc.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xA6:    // LDX (Zero Page)
        ix.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     ix.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, ix.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xA7:    //
        DebugLog.log( 3, "op [A7] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xA8:    // TAY (Implied)
        iy.setValue( acc.getValue() );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     iy.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xA9:    // LDA (Immediate)
        acc.setValue( mainMem.readNoMessage( pc.getValue() + 1 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xAA:    // TAX (Implied)
        ix.setValue( acc.getValue() );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     ix.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xAB:    //
        DebugLog.log( 3, "op [AB] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xAC:    // LDY (Absolute)
        iy.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     iy.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, iy.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xAD:    // LDA (Absolute)
        acc.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xAE:    // LDX (Absolute)
        ix.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     ix.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, ix.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xAF:    //
        DebugLog.log( 3, "op [AF] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xB0:    // BCS (Relative)
        if( willBranch( curOp ) )
          pc.plusEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + NesOpCodeTable.length[curOp] );
        else
          pc.plusEquals( NesOpCodeTable.length[curOp] );  
        break;
      case 0xB1:    // LDA (Indirect,Y)
        acc.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xB2:    //
        DebugLog.log( 3, "op [B2] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xB3:    //
        DebugLog.log( 3, "op [B3] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xB4:    // LDY (Zero Page,X)
        iy.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     iy.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, iy.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xB5:    // LDA (Zero Page,X)
        acc.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xB6:    // LDX (Zero Page,Y)
        ix.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     ix.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, ix.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xB7:    //
        DebugLog.log( 3, "op [B7] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xB8:    // CLV (Implied)
        ps.setBitOff( CpuRegister.PS_OVERFLOW_FLAG );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xB9:    // LDA (Absolute,Y)
        acc.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xBA:    // TSX (Implied)
        ix.setValue( sp.getValue() );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG, ix.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, ix.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xBB:    //
        DebugLog.log( 3, "op [BB] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xBC:    // LDY (Absolute,X)
        iy.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     iy.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, iy.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xBD:    // LDA (Absolute,X)
        acc.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xBE:    // LDX (Absolute,Y)
        ix.setValue( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     ix.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, ix.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xBF:    //
        DebugLog.log( 3, "op [BF] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xC0:    // CPY (Immediate)
        origVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal = (iy.getValue() - origVal) & 0x0000FFFF; 
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    iy.getValue() >= origVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xC1:    // CMP (Indirect,X)
        tempVal = acc.getValue() - getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= tempVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xC2:    //
        DebugLog.log( 3, "op [C2] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xC3:    //
        DebugLog.log( 3, "op [C3] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xC4:    // CPY (Zero Page)
        origVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal = (iy.getValue() - origVal) & 0x0000FFFF;
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    iy.getValue() >= origVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xC5:    // CMP (Zero Page)
        origVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal = (acc.getValue() - origVal) & 0x0000FFFF;
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= origVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xC6:    // DEC (Zero Page)
        addr = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        mainMem.writeNoMessage( addr, mainMem.readNoMessage( addr ) - 1 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     mainMem.readNoMessage(addr) == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, mainMem.readNoMessage(addr) & MathHelper.bitMask[7] );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xC7:    //
        DebugLog.log( 3, "op [C7] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xC8:    // INY (Implied)
        iy.increment();
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     iy.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, iy.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xC9:    // CMP (Immediate)
        origVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal = (acc.getValue() - origVal) & 0x0000FFFF;
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= origVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xCA:    // DEX (Implied)
        ix.decrement();
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     ix.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, ix.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xCB:    //
        DebugLog.log( 3, "op [CB] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xCC:    // CPY (Absolute)
        origVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        tempVal = (iy.getValue() - origVal) & 0x0000FFFF;
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    iy.getValue() >= origVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xCD:    // CMP (Absolute)
        origVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        tempVal = (acc.getValue() - origVal) & 0x0000FFFF;
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= origVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xCE:    // DEC (Absolute)
        addr = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        mainMem.writeNoMessage( addr, mainMem.readNoMessage( addr ) - 1 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     mainMem.readNoMessage(addr) == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, mainMem.readNoMessage(addr) & MathHelper.bitMask[7] );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xCF:    //
        DebugLog.log( 3, "op [CF] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xD0:    // BNE (Relative)
        if( willBranch( curOp ) )
          pc.plusEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + NesOpCodeTable.length[curOp] );
        else
          pc.plusEquals( NesOpCodeTable.length[curOp] );  
        break;
      case 0xD1:    // CMP (Indirect,Y)
        tempVal = acc.getValue() - getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= tempVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xD2:    //
        DebugLog.log( 3, "op [D2] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xD3:    //
        DebugLog.log( 3, "op [D3] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xD4:    //
        DebugLog.log( 3, "op [D4] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xD5:    // CMP (Zero Page,X)
        origVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal = (acc.getValue() - origVal) & 0x0000FFFF;
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= origVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xD6:    // DEC (Zero Page,X)
        addr = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        mainMem.writeNoMessage( addr, mainMem.readNoMessage( addr ) - 1 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     mainMem.readNoMessage(addr) == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, mainMem.readNoMessage(addr) & MathHelper.bitMask[7] );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xD7:    //
        DebugLog.log( 3, "op [D7] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xD8:    // CLD (Implied)
        ps.setBitOff( CpuRegister.PS_DECIMAL_MODE );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xD9:    // CMP (Absolute,Y)
        tempVal = acc.getValue() - getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= tempVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xDA:    //
        DebugLog.log( 3, "op [DA] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xDB:    //
        DebugLog.log( 3, "op [DB] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xDC:    //
        DebugLog.log( 3, "op [DC] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xDD:    // CMP (Absolute,X)
        tempVal = acc.getValue() - getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= tempVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xDE:    // DEC (Absolute,X)
        addr = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        mainMem.writeNoMessage( addr, mainMem.readNoMessage( addr ) - 1 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     mainMem.readNoMessage(addr) == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, mainMem.readNoMessage(addr) & MathHelper.bitMask[7] );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xDF:    //
        DebugLog.log( 3, "op [DF] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xE0:    // CPX (Immediate)
        origVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal = (ix.getValue() - origVal) & 0x0000FFFF;
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    ix.getValue() >= origVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xE1:    // SBC (Indirect,X)
        tempVal = acc.getValue() - getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) - (~(ps.getBitValue( CpuRegister.PS_CARRY_FLAG )) & MathHelper.bitMask[0]);
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= 0 );        
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xE2:    //
        DebugLog.log( 3, "op [E2] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xE3:    //
        DebugLog.log( 3, "op [E3] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xE4:    // CPX (Zero Page)
        origVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        tempVal = (ix.getValue() - origVal) & 0x0000FFFF;
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    ix.getValue() >= origVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xE5:    // SBC (Zero Page)
        tempVal = acc.getValue() - getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) - (~(ps.getBitValue( CpuRegister.PS_CARRY_FLAG )) & MathHelper.bitMask[0]);
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= 0 );        
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xE6:    // INC (Zero Page)
        addr = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        mainMem.writeNoMessage( addr, mainMem.readNoMessage( addr ) + 1 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     mainMem.readNoMessage(addr) == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, mainMem.readNoMessage(addr) & MathHelper.bitMask[7] );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xE7:    //
        DebugLog.log( 3, "op [E7] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xE8:    // INX (Implied)
        ix.increment();
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     ix.getValue() == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, ix.getBitValue(MathHelper.bitMask[7]) );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xE9:    // SBC (Immediate)
        tempVal = acc.getValue() - getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) - (~(ps.getBitValue( CpuRegister.PS_CARRY_FLAG )) & MathHelper.bitMask[0]);
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= 0 );        
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xEA:    // NOP (Implied)
        // Do nothing!
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xEB:    //
        DebugLog.log( 3, "op [EB] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xEC:    // CPX (Absolute)
        origVal = getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        tempVal = (ix.getValue() - origVal) & 0x0000FFFF;
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    ix.getValue() >= origVal );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, tempVal & MathHelper.bitMask[7] );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     tempVal == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xED:    // SBC (Absolute)
        tempVal = acc.getValue() - getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) - (~(ps.getBitValue( CpuRegister.PS_CARRY_FLAG )) & MathHelper.bitMask[0]);
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= 0 );        
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xEE:    // INC (Absolute)
        addr = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        mainMem.writeNoMessage( addr, mainMem.readNoMessage( addr ) + 1 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     mainMem.readNoMessage(addr) == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, mainMem.readNoMessage(addr) & MathHelper.bitMask[7] );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xEF:    //
        DebugLog.log( 3, "op [EF] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xF0:    // BEQ (Relative)
        if( willBranch( curOp ) )
          pc.plusEquals( getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) + NesOpCodeTable.length[curOp] );
        else
          pc.plusEquals( NesOpCodeTable.length[curOp] );  
        break;
      case 0xF1:    // SBC (Indirect,Y)
        tempVal = acc.getValue() - getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) - (~(ps.getBitValue( CpuRegister.PS_CARRY_FLAG )) & MathHelper.bitMask[0]);
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= 0 );        
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xF2:    //
        DebugLog.log( 3, "op [F2] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xF3:    //
        DebugLog.log( 3, "op [F3] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xF4:    //
        DebugLog.log( 3, "op [F4] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xF5:    // SBC (Zero Page,X)
        tempVal = acc.getValue() - getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 ) - (~(ps.getBitValue( CpuRegister.PS_CARRY_FLAG )) & MathHelper.bitMask[0]);
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= 0 );        
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xF6:    // INC (Zero Page,X)
        addr = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), 0 );
        mainMem.writeNoMessage( addr, mainMem.readNoMessage( addr ) + 1 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     mainMem.readNoMessage(addr) == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, mainMem.readNoMessage(addr) & MathHelper.bitMask[7] );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xF7:    //
        DebugLog.log( 3, "op [F7] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xF8:    // SED (Implied)
        ps.setBitOn( CpuRegister.PS_DECIMAL_MODE );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xF9:    // SBC (Absolute,Y)
        tempVal = acc.getValue() - getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) - (~(ps.getBitValue( CpuRegister.PS_CARRY_FLAG )) & MathHelper.bitMask[0]);
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= 0 );        
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xFA:    //
        DebugLog.log( 3, "op [FA] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xFB:    //
        DebugLog.log( 3, "op [FB] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xFC:    //
        DebugLog.log( 3, "op [FC] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xFD:    // SBC (Absolute,X)
        tempVal = acc.getValue() - getOpValue( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) ) - (~(ps.getBitValue( CpuRegister.PS_CARRY_FLAG )) & MathHelper.bitMask[0]);
        ps.setBitValue( CpuRegister.PS_OVERFLOW_FLAG, acc.getBitValue( MathHelper.bitMask[7]) != ((tempVal & MathHelper.bitMask[7]) >> 7) );
        acc.setValue( tempVal );
        ps.setBitValue( CpuRegister.PS_CARRY_FLAG,    acc.getValue() >= 0 );        
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, acc.getBitValue(MathHelper.bitMask[7]) );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     acc.getValue() == 0 );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xFE:    // INC (Absolute,X)
        addr = getOpAddress( curOp, mainMem.readNoMessage( pc.getValue() + 1 ), mainMem.readNoMessage( pc.getValue() + 2 ) );
        mainMem.writeNoMessage( addr, mainMem.readNoMessage( addr ) + 1 );
        ps.setBitValue( CpuRegister.PS_ZERO_FLAG,     mainMem.readNoMessage(addr) == 0 );
        ps.setBitValue( CpuRegister.PS_NEGATIVE_FLAG, mainMem.readNoMessage(addr) & MathHelper.bitMask[7] );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
      case 0xFF:    //
        DebugLog.log( 3, "op [FF] - unknown" );
        pc.plusEquals( NesOpCodeTable.length[curOp] );
        break;
    }

    availCycles -= needCycles;
    needCycles = getNextInstructionCycles();
  }


  /**
   * Gets the parameter value based on the OpCode.
   * @param inOpCode The OpCode whose mode will be used to determine the value.
   * @param byte1 The first byte of the param used to lookup the full value.
   * @param byte2 The second byte of the param used to lookup the full value.
   * @return The final operation parameter value.
   * @see com.monkeycoder.monkeynes.hardware.NesCpu#getOpAddress(int,int,int)   
   */
  public int getOpValue( int inOpCode, int byte1, int byte2 )
  {
    int addr1    = 0;
    int addr2    = 0;
    int preval   = 0;

    switch( NesOpCodeTable.mode[inOpCode] )
    {
      // Don't use this function for these modes! -------------------------
      case NesOpCodeTable.MODE_IMPLIED:          // Nothing to do!
      case NesOpCodeTable.MODE_ACCUMULATOR:      // Nothing to do!
        return 0;

      // You can use it for this one, but it's silly.
      case NesOpCodeTable.MODE_IMMEDIATE:        // byte1 is the value
        return byte1;
      // ------------------------------------------------------------------

      case NesOpCodeTable.MODE_ZERO_PAGE:        // Address = $00xx (xx = byte1)
        return mainMem.readNoMessage( byte1 );

      case NesOpCodeTable.MODE_ZERO_PAGE_X:      // Address = byte1 + ix
        return mainMem.readNoMessage( (byte1 + ix.getValue()) & 0x000000FF );

      case NesOpCodeTable.MODE_ZERO_PAGE_Y:      // Address = byte1 + iy
        return mainMem.readNoMessage( (byte1 + iy.getValue()) & 0x000000FF );

      case NesOpCodeTable.MODE_ABSOLUTE:         // Address = $yyxx  (xx = byte1, yy = byte2)
        return mainMem.readNoMessage( ((byte2 << 8) | (byte1)) );

      case NesOpCodeTable.MODE_ABSOLUTE_X:       // Address = $yyxx + ix  (xx = byte1, yy = byte2)
        return mainMem.readNoMessage( ((byte2 << 8) | (byte1)) + ix.getValue() );

      case NesOpCodeTable.MODE_ABSOLUTE_Y:       // Address = $yyxx + iy  (xx = byte1, yy = byte2)
        return mainMem.readNoMessage( ((byte2 << 8) | (byte1)) + iy.getValue() );

      case NesOpCodeTable.MODE_INDIRECT:         // Address = value at $yyxx  (xx = byte1, yy = byte2)
        preval = ((byte2 << 8) | (byte1));
        addr1 = mainMem.readNoMessage( preval );
        preval = ((byte2 << 8) | ((byte1 + 1) & 0x000000FF));
        addr2 = MathHelper.getWrappedByte( mainMem.readNoMessage( preval ) );
        return ((addr2 << 8) | (addr1));

      case NesOpCodeTable.MODE_INDIRECT_X:       // Address = value at ($00xx + ix)  (xx = byte1)
        addr1 = mainMem.readNoMessage( byte1 + ix.getValue() );
        addr2 = MathHelper.getWrappedByte( mainMem.readNoMessage( byte1 + ix.getValue() + 1 ) );
        return mainMem.readNoMessage( ((addr2 << 8) | (addr1)) );

      case NesOpCodeTable.MODE_INDIRECT_Y:       // Address = (value at $00xx) + iy  (xx = byte1)
        addr1  = mainMem.readNoMessage( byte1 );
        addr2  = mainMem.readNoMessage( byte1 + 1 );
        preval = mainMem.readNoMessage( ((addr2 << 8) | (addr1)) );
        return mainMem.readNoMessage( preval + iy.getValue() );

      case NesOpCodeTable.MODE_RELATIVE:         // byte1 is the relative number
        if( (byte1 & MathHelper.bitMask[7]) == MathHelper.bitMask[7] )
          return -(byte1 & 0x0000007F);
        else
          return (byte1 & 0x0000007F);
    }

    return 0;
  }
  
  /**
   * Gets the address at which the parameter value is located.  This method acts 
   * as the memory pointer counterpart to the function getOpValue().  (As that
   * makes sense based on OpCode mode.)
   * @param inOpCode The OpCode whose mode will be used to determine the value.
   * @param byte1 The first byte of the param used to lookup the full value.
   * @param byte2 The second byte of the param used to lookup the full value.
   * @return The address of the final operation parameter value.
   * @see com.monkeycoder.monkeynes.hardware.NesCpu#getOpValue(int,int,int)
   */  
  public int getOpAddress( int inOpCode, int byte1, int byte2 )
  {
    int addr1    = 0;
    int addr2    = 0;
    int preval   = 0;

    switch( NesOpCodeTable.mode[inOpCode] )
    {
      // Don't use this function for these modes! -------------------------
      case NesOpCodeTable.MODE_IMPLIED:          // Nothing to do!
      case NesOpCodeTable.MODE_ACCUMULATOR:      // Nothing to do!
        return 0;

      // You can use it for this one, but it's silly.
      case NesOpCodeTable.MODE_IMMEDIATE:        // byte1 is the value
      case NesOpCodeTable.MODE_ZERO_PAGE:        // Address = $00xx (xx = byte1)
        return byte1;
      // ------------------------------------------------------------------

      case NesOpCodeTable.MODE_ZERO_PAGE_X:      // Address = byte1 + ix
        return (byte1 + ix.getValue()) & 0x000000FF;

      case NesOpCodeTable.MODE_ZERO_PAGE_Y:      // Address = byte1 + iy
        return (byte1 + iy.getValue()) & 0x000000FF;

      case NesOpCodeTable.MODE_ABSOLUTE:         // Address = $yyxx  (xx = byte1, yy = byte2)
        return ((byte2 << 8) | (byte1));

      case NesOpCodeTable.MODE_ABSOLUTE_X:       // Address = $yyxx + ix  (xx = byte1, yy = byte2)
        return ((byte2 << 8) | (byte1)) + ix.getValue();

      case NesOpCodeTable.MODE_ABSOLUTE_Y:       // Address = $yyxx + iy  (xx = byte1, yy = byte2)
        return ((byte2 << 8) | (byte1)) + iy.getValue();

      case NesOpCodeTable.MODE_INDIRECT:         // Address = value at $yyxx  (xx = byte1, yy = byte2)
        preval = ((byte2 << 8) | (byte1));
        addr1 = mainMem.readNoMessage( preval );
        preval = ((byte2 << 8) | ((byte1 + 1) & 0x000000FF));
        addr2 = MathHelper.getWrappedByte( mainMem.readNoMessage( preval ) );
        return ((addr2 << 8) | (addr1));

      case NesOpCodeTable.MODE_INDIRECT_X:       // Address = value at ($00xx + ix)  (xx = byte1)
        addr1 = mainMem.readNoMessage( byte1 + ix.getValue() );
        addr2 = MathHelper.getWrappedByte( mainMem.readNoMessage( byte1 + ix.getValue() + 1 ) );
        return ((addr2 << 8) | (addr1));

      case NesOpCodeTable.MODE_INDIRECT_Y:       // Address = (value at $00xx) + iy  (xx = byte1)
        addr1  = mainMem.readNoMessage( byte1 );
        addr2  = mainMem.readNoMessage( byte1 + 1 );
        preval = mainMem.readNoMessage( ((addr2 << 8) | (addr1)) );
        return preval + iy.getValue();

      case NesOpCodeTable.MODE_RELATIVE:         // byte1 is the relative number
        if( (byte1 & MathHelper.bitMask[7]) == MathHelper.bitMask[7] )
          return -(byte1 & 0x0000007F);
        else
          return (byte1 & 0x0000007F);
    }

    return 0;
  }
     
  
  /**
   * Determines whether or not an operation is going to cross a page boundry.
   * @param inOpCode The OpCode whose mode will be used to determine the boundry crossing effect.
   * @param byte1 The first byte of the param used.
   * @param byte2 The second byte of the param used.
   * @return Whether or not the operation crosses a page boundry and requires additional cycles.
   */  
  public boolean crossesPageBoundry( int inOpCode, int byte1, int byte2 )
  {
    int curOpPage    = 0;
    int baseAddrPage = 0;
    int tempPage     = 0;
    int addr1        = 0;
    int addr2        = 0;
    int preval       = 0;    

    switch( NesOpCodeTable.mode[inOpCode] )
    {
      case NesOpCodeTable.MODE_ABSOLUTE_X:
        baseAddrPage = byte2;
        addr1 = ((byte2 << 8) | (byte1)) + ix.getValue();
        tempPage = (addr1 >> 8) & 0x000000FF;
        return (baseAddrPage != tempPage);
      
      case NesOpCodeTable.MODE_ABSOLUTE_Y:
        baseAddrPage = byte2;
        addr1 = ((byte2 << 8) | (byte1)) + iy.getValue();
        tempPage = (addr1 >> 8) & 0x000000FF;
        return (baseAddrPage != tempPage);
      
      case NesOpCodeTable.MODE_INDIRECT_Y:
        addr1  = mainMem.readNoMessage( byte1 );
        addr2  = mainMem.readNoMessage( byte1 + 1 );
        preval = mainMem.readNoMessage( ((addr2 << 8) | (addr1)) );
        baseAddrPage = (preval >> 8) & 0x000000FF;
        tempPage = (((preval + iy.getValue()) >> 8) & 0x000000FF);
        return (baseAddrPage != tempPage);
      
      case NesOpCodeTable.MODE_RELATIVE:   
        curOpPage = (pc.getValue() >> 8) & 0x000000FF;
        if( willBranch( inOpCode ) )
          preval = getOpValue( inOpCode, byte1, byte2 );            // The relative jump size
        else
          preval = NesOpCodeTable.length[inOpCode];                 // The size of the Op
        tempPage = ((pc.getValue() + preval) >> 8) & 0x000000FF;    // Target page
        return (tempPage != curOpPage);
  
      // Page Boundries Don't Matter for these Modes! ---------------------
      case NesOpCodeTable.MODE_IMPLIED:
      case NesOpCodeTable.MODE_ACCUMULATOR:
      case NesOpCodeTable.MODE_IMMEDIATE:
      case NesOpCodeTable.MODE_ZERO_PAGE:
      case NesOpCodeTable.MODE_ZERO_PAGE_X:
      case NesOpCodeTable.MODE_ZERO_PAGE_Y:
      case NesOpCodeTable.MODE_ABSOLUTE:
      case NesOpCodeTable.MODE_INDIRECT:
      case NesOpCodeTable.MODE_INDIRECT_X:
      default:
        return false;
      // ------------------------------------------------------------------
    }
  }
  
  public boolean willBranch( int inOpCode )
  {
    switch( inOpCode )
    {
      case 0x10:  // BPL (Relative)
        return (ps.getBitValue( CpuRegister.PS_NEGATIVE_FLAG ) == 0);
      case 0x30:  // BMI (Relative)
        return (ps.getBitValue( CpuRegister.PS_NEGATIVE_FLAG ) == 1);
      case 0x50:  // BVC (Relative)
        return (ps.getBitValue( CpuRegister.PS_OVERFLOW_FLAG ) == 0);
      case 0x70:  // BVS (Relative)
        return (ps.getBitValue( CpuRegister.PS_OVERFLOW_FLAG ) == 1);
      case 0x90:  // BCC (Relative)
        return (ps.getBitValue( CpuRegister.PS_CARRY_FLAG ) == 0 );
      case 0xB0:  // BCS (Relative)
        return (ps.getBitValue( CpuRegister.PS_CARRY_FLAG ) == 1 );
      case 0xD0:  // BNE (Relative)
        return (ps.getBitValue( CpuRegister.PS_ZERO_FLAG ) == 0 );
      case 0xF0:  // BEQ (Relative)
        return (ps.getBitValue( CpuRegister.PS_ZERO_FLAG ) == 1 );
      default:
        return false;
    }
  }

  /**
    * For use by the OpCode Test Shell.  This function will likely have trouble with 
    * jump and brance instructions.  The number of cycles needed to complete the 
    * requested operation are automatically and instantly granted.  The operation binary
    * data is loaded at $8000-$8002 and will destroy whatever data is already there in
    * the process.  The data bytes are expected to be in exactly the same order as 
    * they would appear in a ROM.
    * @param inOpCode The raw OpCode value.
    * @param inByte1 The first byte of raw parameter data.
    * @param inByte2 The second byte of raw parameter data.
    * @return The number of cycles the operation took.
    * @see com.monkeycoder.monkeynes.gui.OpCodeTestShellGui
    */
  public int testExecuteInstruction( int inOpCode, int inByte1, int inByte2 )
  {
    int cycles = 0;
    
    pc.setValue( 0 );
    mainMem.writeNoMessage( 0, inOpCode );
    mainMem.writeNoMessage( 1, inByte1 );
    mainMem.writeNoMessage( 2, inByte2 );
        
    needCycles  = getNextInstructionCycles();
    availCycles = needCycles;
    cycles      = needCycles;
    
    executeNextInstruction();
    
    return cycles;
  }
  
  /**
    * For use by the OpCode Test Shell.  This function does not alter any memory directly
    * and will simply execute the next instruction as found in memory at the point currently
    * referenced by the program counter.
    * @return An ExecutedInstructionInfo object containing info about what was just executed.
    * @see com.monkeycoder.monkeynes.gui.OpCodeTestShellGui
    */
  public ExecutedInstructionInfo testStepInstruction()
  {
    ExecutedInstructionInfo eii = null;
    
    needCycles  = getNextInstructionCycles();
    availCycles = needCycles;
    
    // Grab info about what's about to happen
    int opAddr = getProgramCounterValue();
    int opCode = mainMem.readNoMessage( opAddr );
    switch( NesOpCodeTable.length[opCode] )
    {
      case 1:
        eii = new ExecutedInstructionInfo( needCycles, opCode, 0, 0 );
        break;
      case 2:
        eii = new ExecutedInstructionInfo( needCycles, opCode, mainMem.readNoMessage( opAddr + 1 ), 0 );
        break;
      case 3:
        eii = new ExecutedInstructionInfo( needCycles, opCode, mainMem.readNoMessage( opAddr + 1 ), mainMem.readNoMessage( opAddr + 2 ) );
        break;
    }
    
    // Actually let the instruction run now that we know all about it
    executeNextInstruction();
    
    return eii; 
  }
  
  
  /**
   * Gets the value of the Program Counter (PC) register.
   * @return The value of the PC register.
   */
  public int getProgramCounterValue()
  {
    return pc.getValue();
  }

  /**
   * Gets the value of the Stack Pointer (SP) register.
   * @return The value of the SP register.
   */
  public int getStackPointerValue()
  {
    return sp.getValue();
  }

  /**
   * Gets the value of the Accumulator (ACC) register.
   * @return The value of the ACC register.
   */
  public int getAccumulatorValue()
  {
    return acc.getValue();
  }

  /**
   * Gets the value of the X-Index (IX) register.
   * @return The value of the IX register.
   */
  public int getXIndexValue()
  {
    return ix.getValue();
  }

  /**
   * Gets the value of the Y-Index (IX) register.
   * @return The value of the IX register.
   */
  public int getYIndexValue()
  {
    return iy.getValue();
  }

  /**
   * Gets the value of the Processor Status (PS) register.
   * @return The value of the PS register.
   */
  public int getProcessorStatusValue()
  {
    return ps.getValue();
  }
  
  /**
   * Sets the value of the Program Counter (PC) register.
   * @return The value of the PC register.
   */
  public void setProgramCounterValue( int inValue )
  {
    pc.setValue( inValue );
  }

  /**
   * Sets the value of the Stack Pointer (SP) register.
   * @return The value of the SP register.
   */
  public void setStackPointerValue( int inValue )
  {
    sp.setValue( inValue );
  }

  /**
   * Sets the value of the Accumulator (ACC) register.
   * @return The value of the ACC register.
   */
  public void setAccumulatorValue( int inValue )
  {
    acc.setValue( inValue );
  }

  /**
   * Sets the value of the X-Index (IX) register.
   * @return The value of the IX register.
   */
  public void setXIndexValue( int inValue )
  {
    ix.setValue( inValue );
  }

  /**
   * Sets the value of the Y-Index (IX) register.
   * @return The value of the IX register.
   */
  public void setYIndexValue( int inValue )
  {
    iy.setValue( inValue );
  }

  /**
   * Sets the value of the Processor Status (PS) register.
   * @return The value of the PS register.
   */
  public void setProcessorStatusValue( int inValue )
  {
    ps.setValue( inValue );
  }
  
  /**
   * Gets the number of cycles the CPU is waiting for before completing the next operation.
   * @return The number of cycles the CPU requires to complete the next operation.
   */
  public int getNeedCycles()
  {
    return needCycles;
  }

  /**
   * Gets the number of cycles the CPU has waiting to be used by the next operation.
   * @return The number of cycles the CPU has waiting to be used by the next operation.
   */  
  public int getAvailCycles()
  {
    return availCycles;
  }    
}