package com.monkeycoder.monkeynes.hardware;

/**
 * The NesRom class represents a physical NES game cartrige.  It contains all of the 
 * program and graphics data, information about the structure of the data and about 
 * the presence or abscence of hardware used to manipulate the data.  At the moment
 * the only supported file format that can be used to create a NesRom is the INES
 * format.  This is due to the completeness of the format as well as for the fact 
 * that virtually no other formats even exist in circulation anymore.
 * <br><br> 
 * Originally part of the MonkeyNES project.
 *
 * @author David Stephens
 */
public class NesRom
{
  // 40k                    = 40960
  // supermariobros.nes     = 40976
  // NES format header size = 16 bytes;
   
  public static final int ROM_FORMAT_UNKNOWN = 0;   
  public static final int ROM_FORMAT_INES    = 1;
  
  private int[]  romFormatHeaderSize   = { 0, 16 };
  
  private String        romFileName  = null;
  private int           romFormat    = ROM_FORMAT_UNKNOWN;
  private InesRomHeader inesHead     = null;
  private byte[]        headerData   = null;
  private int[]         romData      = null;
  
  /**
   * Creates an empty NesRom.
   */
  public NesRom()
  {
    romFileName  = null;              
    romFormat    = ROM_FORMAT_UNKNOWN;
    inesHead     = null;
    headerData   = null;              
    romData      = null;              
  }
  
  /**
   * Sets the filename from which the data for this ROM was loaded.
   * @param inRomFileName The filename.
   */
  public void setRomFileName( String inRomFileName )
  {
    romFileName = inRomFileName;
  }

  /**
   * Sets the ROM format.
   * @param inRomFormat The format identifier.
   */  
  public void setRomFormat( int inRomFormat )
  {
    romFormat = inRomFormat;
  }

  /**
   * Sets the ROM header data.  This data is expected to be identical to the 
   * file header located in the ROM file.  This data is converted into an <code>InesRomHeader</code>.
   * @param inHeaderData An array of bytes from the ROM file representing the header.
   * @see com.monkeycoder.monkeynes.hardware.InesRomHeader
   */  
  public void setHeaderData( byte[] inHeaderData )
  {
    headerData = inHeaderData;
    inesHead   = new InesRomHeader( headerData );
  }

  /**
   * Sets the ROM data.  This data is expected to be identical to the complete data
   * for the game ROM located in the ROM file.
   * @param inRomData An array of bytes from the ROM file representing the game data.
   */  
  public void setRomData( int[] inRomData )
  {
    romData = inRomData;
  }      
  
  /**
   * Returns the file filename from which the data for this ROM was loaded.
   * @return The filename.
   */
  public String getRomFileName()
  {
    return romFileName;
  }

  /**
   * Returns the file format identifier.
   * @return The file format identifier.
   */  
  public int getRomFormat()
  {
    return romFormat;
  }

  /**
   * Returns the <code>InesRomHeader</code> generated when the raw header data
   * was passed in.
   * @return The InesRomHeader.
   * @see com.monkeycoder.monkeynes.hardware.InesRomHeader
   */
  public InesRomHeader getInesRomHeader()
  {
    return inesHead;
  }  

  /**
   * Returns the raw header data as it was passed in.
   * @return The raw header data.
   */  
  public byte[] getHeaderData()
  {
    return headerData;
  }

  /**
   * Returns the raw game data as it was passed in.
   * @return The raw game data.
   */  
  public int[] getRomData()
  {
    return romData;
  }
  
  /**
   * Returns the size of the header data in bytes.
   * @return The size of the header data in bytes.
   */
  public int getRomFormatHeaderSize( int inRomFormat )
  {
    return romFormatHeaderSize[inRomFormat];
  }
  
  /** 
   * Returns the data from the ROM representing the requested page 
   * of PRG-ROM.  Pages are numbered starting with 0.
   * @param pageNum The requested page of PRG-ROM.
   * @return The page of PRG-ROM.
   */
  public int[] getPrgRomPage( int pageNum )
  {
    if( pageNum < 0 || pageNum >= inesHead.getPrgRomPageCount() )
      return null;
    
    int start = pageNum * InesRomHeader.PRG_PAGE_SIZE;
    int stop  = start + InesRomHeader.PRG_PAGE_SIZE - 1;
    return read( start, stop );
  }

  /** 
   * Returns the data from the ROM representing the requested page 
   * of CHR-ROM.  Pages are numbered starting with 0.
   * @param pageNum The requested page of CHR-ROM.
   * @return The page of CHR-ROM.
   */
  public int[] getChrRomPage( int pageNum )
  {
    if( pageNum < 0 || pageNum >= inesHead.getChrRomPageCount() )
      return null;
    
    int start = inesHead.getPrgRomSize() + (pageNum * InesRomHeader.CHR_PAGE_SIZE);
    int stop  = start + InesRomHeader.CHR_PAGE_SIZE - 1;
    return read( start, stop );
  }   
 
  /** 
   * Reads any single byte of data from the ROM data.  There are
   * no hints provided about what type of data might be at the requested 
   * address.
   * @param addr The address of the requested data.
   * @return The byte of data requested.
   */
  public int read( int addr )
  {
    return romData[addr];
  }

  /** 
   * Reads a range of bytes from the ROM data.  There are no hints 
   * provided about what type of data might be at the requested 
   * address range.  The range is inclusive.
   * @param startAddr The starting address of the requested data.
   * @param stopAddr The ending address of the requested data.
   * @return The bytes of data requested.
   */
  public int[] read( int startAddr, int stopAddr )
  {
    int j = 0;
    int[] temp = new int[stopAddr - startAddr + 1];
    
    for( int i = startAddr; i <= stopAddr; i++ )
    {
      temp[j] = read(i);
      j++;
    }
    
    return temp;
  }
  
  /**
   * Updates the ROM data at the specified address with a new value.
   * @param addr The address to write.
   * @param val The value to write.
   */
  public void write( int addr, int val )
  {
    romData[addr] = val;
  }

  /**
   * Updates the ROM data starting at the specified address with a new values.
   * The entire incoming data block is written.
   * @param startAddr The address at which to start writing.
   * @param data The incoming data block to write.
   */  
  public void write( int startAddr, int[] data )
  {
    // writes the whole input buffer
    for( int i = 0; i < data.length; i++ )
      write( startAddr + i, data[i] );
  }

  /**
   * Updates the ROM data starting at the specified address with a new values.
   * The incoming data block is written from the beginning but only until the 
   * specified number of bytes has been written.
   * @param startAddr The address at which to start writing.
   * @param data The incoming data block to write.
   * @param len The number of bytes to write.
   * @see com.monkeycoder.monkeynes.hardware.NesVideoMemory#writeNoMessage(int,int[],int)
   */
  public void write( int startAddr, int[] data, int len )
  {
    // writes up to len of the input buffer
    for( int i = 0; i < len; i++ )
      write( startAddr + i, data[i] );
  }
 
  /**
   * Creates a text readout detailing all known data for this ROM.
   * @return A String containing details on everything known about this ROM.
   */
  public String toString()
  {
    StringBuffer sb = new StringBuffer();

    sb.append( "    File: " );
    sb.append( romFileName );
    sb.append( "\n    Format:        " );
    sb.append( romFormat );
    sb.append( "\n    Header Size:   " );
    sb.append( headerData.length );
    sb.append( "\n    ROM Size:      " );
    sb.append( romData.length );
    sb.append( "\n    PRG-ROM:       " );
    sb.append( inesHead.getPrgRomSize() );
    sb.append( " [" + inesHead.getPrgRomPageCount() + " pages]" );
    sb.append( "\n    CHR-ROM:       " );
    sb.append( inesHead.getChrRomSize() );
    sb.append( " [" + inesHead.getChrRomPageCount() + " pages]" );
    sb.append( "\n    Mapper:        " );
    sb.append( inesHead.getMapperNumber() );
    sb.append( "\n    4-Screen VRAM: " );
    sb.append( inesHead.isFourScreenVram() );
    sb.append( "\n    Trainer:       " );
    sb.append( inesHead.isTrainerPresent() );
    sb.append( "\n    Save RAM:      " );
    sb.append( inesHead.isSramEnabled() );
    sb.append( "\n    Vert Mirror:   " );
    sb.append( inesHead.isVertMirror() );
    sb.append( "\n    Horz Mirror:   " );
    sb.append( inesHead.isHorzMirror() );
    
    return sb.toString();
  }


}