package com.monkeycoder.monkeynes.hardware;

import com.monkeycoder.monkeynes.base.*;

import java.util.*;

/**
 * This is a sprite memory bank for the NES.  It has only 256 bytes and is accessible 
 * via DMA registers.  This class is observable and will send all varieties
 * of access messages to all its observers.  There is also a full second set of
 * access functions that do not send out messages.  These are intended for use 
 * by speed intensive pieces of code where messages aren't needed.  Please note
 * that each byte of memory is actual a 32-bit integer and will be automatically
 * limited to 8-bits internally.
 * <br><br> 
 * Originally part of the MonkeyNES project.
 *
 * @author David Stephens
 */
public class NesSpriteMemory extends Observable
{
  private int[] memory = new int[0x00000100];  // 256 bytes
    
  /**
   * Creates a new NesSpriteMemory initialized to all 0's.
   */
  public NesSpriteMemory()
  {
    // Do nothing.
  }
  
  /**
   * Reinitializes the memory bank to all zeros.
   */  
  public void clear()
  {
    memory = new int[0x00000100];  // 256 bytes
    
    setChanged();
    notifyObservers( new MemoryAccessMessage( MemoryAccessMessage.MEM_SPRITE, MemoryAccessMessage.TYPE_CLEAR ) );
  }
  
  /**
   * Reads the memory at the requested address.
   * This function sends out a MemoryAccessMessage indicating what was accessed.
   * @param addr The address to read.
   * @return The value found in memory at the requested address.
   * @see com.monkeycoder.monkeynes.hardware.NesSpriteMemory#readNoMessage(int)
   */  
  public int read( int addr )
  {
    setChanged();
    notifyObservers( new MemoryAccessMessage( MemoryAccessMessage.MEM_SPRITE, MemoryAccessMessage.TYPE_READ, addr ) );
    
    return memory[addr];
  }

  /**
   * Reads the memory at the requested address.
   * This function does not send out a MemoryAccessMessage.
   * @param addr The address to read.
   * @return The value found in memory at the requested address.
   * @see com.monkeycoder.monkeynes.hardware.NesSpriteMemory#read(int)
   */
  public int readNoMessage( int addr )
  {
    return memory[addr];
  }

  /**
   * Reads 2-bytes of memory starting at the requested address.
   * This function sends out a MemoryAccessMessage indicating what was accessed.
   * @param addr The address of the start of the word to read.
   * @return The value found in memory at the requested address.
   * @see com.monkeycoder.monkeynes.hardware.NesSpriteMemory#readWordNoMessage(int)
   */
  public int readWord( int addr )
  {
    setChanged();
    notifyObservers( new MemoryAccessMessage( MemoryAccessMessage.MEM_SPRITE, MemoryAccessMessage.TYPE_READ_WORD, addr, addr + 1 ) );
    
    return (memory[MathHelper.getWrappedAddress(addr + 1)] << 8) | memory[MathHelper.getWrappedAddress(addr)];
  }

  /**
   * Reads 2-bytes of memory starting at the requested address.
   * This function does not send out a MemoryAccessMessage.
   * @param addr The address of the start of the word to read.
   * @return The value found in memory at the requested address.
   * @see com.monkeycoder.monkeynes.hardware.NesSpriteMemory#readWord(int)
   */  
  public int readWordNoMessage( int addr )
  {
    return (memory[MathHelper.getWrappedAddress(addr + 1)] << 8) | memory[MathHelper.getWrappedAddress(addr)];
  }  

  /**
   * Reads the memory at the requested address range.  The access is 
   * inclusive to both ends of the range.
   * This function sends out a MemoryAccessMessage indicating what was accessed.
   * @param startAddr The address at which the read will start.
   * @param stopAddr The address at which the read will stop.
   * @return The value found in memory at the requested address.
   * @see com.monkeycoder.monkeynes.hardware.NesSpriteMemory#readNoMessage(int,int)
   */
  public int[] read( int startAddr, int stopAddr )
  {
    setChanged();
    notifyObservers( new MemoryAccessMessage( MemoryAccessMessage.MEM_SPRITE, MemoryAccessMessage.TYPE_READ_RANGE, startAddr, stopAddr ) );

    int j = 0;
    int[] temp = new int[stopAddr - startAddr + 1];
    
    for( int i = startAddr; i <= stopAddr; i++ )
    {
      temp[j] = readNoMessage(i);
      j++;
    }
    
    return temp;
  }  


  /**
   * Reads the memory at the requested address range.  The access is 
   * inclusive to both ends of the range.
   * This function does not send out a MemoryAccessMessage.
   * @param startAddr The address at which the read will start.
   * @param stopAddr The address at which the read will stop.
   * @return The value found in memory at the requested address.
   * @see com.monkeycoder.monkeynes.hardware.NesSpriteMemory#read(int,int)
   */  
  public int[] readNoMessage( int startAddr, int stopAddr )
  {
    int j = 0;
    int[] temp = new int[stopAddr - startAddr + 1];
    
    for( int i = startAddr; i <= stopAddr; i++ )
    {
      temp[j] = readNoMessage(i);
      j++;
    }
    
    return temp;
  }  
  
  /**
   * Updates the memory at the specified address with a new value.
   * This function sends out a MemoryAccessMessage indicating what was accessed.
   * @param addr The address to write.
   * @param val The value to write.
   * @see com.monkeycoder.monkeynes.hardware.NesSpriteMemory#writeNoMessage(int,int)
   */ 
  public void write( int addr, int val )
  {
    memory[addr] = (val & 0x000000FF);

    setChanged();
    notifyObservers( new MemoryAccessMessage( MemoryAccessMessage.MEM_SPRITE, MemoryAccessMessage.TYPE_WRITE, addr ) );
  }

  /**
   * Updates the memory at the specified address with a new value.
   * This function does not send out a MemoryAccessMessage.
   * @param addr The address to write.
   * @param val The value to write.
   * @see com.monkeycoder.monkeynes.hardware.NesSpriteMemory#write(int,int)
   */  
  public void writeNoMessage( int addr, int val )
  {
    memory[addr] = (val & 0x000000FF);
  }  

  /**
   * Updates the memory starting at the specified address with a new values.
   * The entire incoming data block is written.
   * This function sends out a MemoryAccessMessage indicating what was accessed.
   * @param startAddr The address at which to start writing.
   * @param data The incoming data block to write.
   * @see com.monkeycoder.monkeynes.hardware.NesSpriteMemory#writeNoMessage(int,int[])
   */  
  public void write( int startAddr, int[] data )
  {
    // writes the whole input buffer
    for( int i = 0; i < data.length; i++ )
      writeNoMessage( startAddr + i, data[i] );

    setChanged();
    notifyObservers( new MemoryAccessMessage( MemoryAccessMessage.MEM_SPRITE, MemoryAccessMessage.TYPE_WRITE_RANGE, startAddr, startAddr + data.length - 1 ) );
  }

  /**
   * Updates the memory starting at the specified address with a new values.
   * The entire incoming data block is written.
   * This function does not send out a MemoryAccessMessage.
   * @param startAddr The address at which to start writing.
   * @param data The incoming data block to write.
   * @see com.monkeycoder.monkeynes.hardware.NesSpriteMemory#write(int,int[])
   */  
  public void writeNoMessage( int startAddr, int[] data )
  {
    // writes the whole input buffer
    for( int i = 0; i < data.length; i++ )
      writeNoMessage( startAddr + i, data[i] );
  }  

  /**
   * Updates the memory starting at the specified address with a new values.
   * The incoming data block is written from the beginning but only until the 
   * specified number of bytes has been written.
   * This function sends out a MemoryAccessMessage indicating what was accessed.
   * @param startAddr The address at which to start writing.
   * @param data The incoming data block to write.
   * @param len The number of bytes to write.
   * @see com.monkeycoder.monkeynes.hardware.NesSpriteMemory#writeNoMessage(int,int[],int)
   */
  public void write( int startAddr, int[] data, int len )
  {
    // writes up to len of the input buffer
    for( int i = 0; i < len; i++ )
      writeNoMessage( startAddr + i, data[i] );

    setChanged();
    notifyObservers( new MemoryAccessMessage( MemoryAccessMessage.MEM_SPRITE, MemoryAccessMessage.TYPE_WRITE_RANGE, startAddr, startAddr + len - 1 ) );
  }

  /**
   * Updates the memory starting at the specified address with a new values.
   * The incoming data block is written from the beginning but only until the 
   * specified number of bytes has been written.
   * This function does not send out a MemoryAccessMessage.
   * @param startAddr The address at which to start writing.
   * @param data The incoming data block to write.
   * @param len The number of bytes to write.
   * @see com.monkeycoder.monkeynes.hardware.NesSpriteMemory#write(int,int[],int)
   */
  public void writeNoMessage( int startAddr, int[] data, int len )
  {
    // writes up to len of the input buffer
    for( int i = 0; i < len; i++ )
      writeNoMessage( startAddr + i, data[i] );
  }  

}