package com.monkeycoder.monkeynes.hardware.mapper;

import com.monkeycoder.monkeynes.hardware.*;

/**
 * This is the abstract superclass of all hardware mappers.  There are 
 * supposedly hundreds of them.
 * <br><br> 
 * Originally part of the MonkeyNES project.
 *
 * @author David Stephens
 */
public abstract class MemoryMapper
{
  protected NesMainMemory  mainMem = null;
  protected NesVideoMemory vidMem  = null;
  protected NesRom         rom     = null;
  
  /**
   * Creates a new MemoryMapper.
   * @param inMainMem Main memory bank accessed by this mapper.
   * @param inVidMem Video memory bank accessed by this mapper.
   * @param inRom ROM accessed by this mapper.
   */
  protected MemoryMapper( NesMainMemory inMainMem, NesVideoMemory inVidMem, NesRom inRom )
  {
    mainMem = inMainMem;
    vidMem  = inVidMem;
    rom     = inRom;
  }

  /**
   * Returns a reference to the main memory bank this mapper is using.
   * @return Main memory bank accessed by this mapper.  
   */
  public NesMainMemory getMainMemory()
  {
    return mainMem;
  }

  /**
   * Returns a reference to the video memory bank this mapper is using.
   * @return Video memory bank accessed by this mapper.  
   */  
  public NesVideoMemory getVideoMemory()
  {
    return vidMem;
  }

  /**
   * Returns a reference to the ROM this mapper is using.
   * @return ROM accessed by this mapper.  
   */  
  public NesRom getRom()
  {
    return rom;
  } 

  /**
   * Converts the data in this mapper to a string.
   * @return A string representing this this mapper.
   */  
  public String toString()
  {
    StringBuffer sb = new StringBuffer();

    sb.append( "[" );
    sb.append( getMapperNumber() );
    sb.append( "] " );
    sb.append( getMapperName() );
    sb.append( " - " );
    sb.append( getMapperDesc() );   
    
    return sb.toString();
  }
  
  /**
   * Returns the mapper number of this mapper.
   * @return The mapper number of this mapper.
   */
  public abstract int           getMapperNumber();
  
  /**
   * Returns the name of this mapper.
   * @return The name of this mapper.
   */
  public abstract String        getMapperName();
  
  /**
   * Returns the description of this mapper.
   * @return The description of this mapper.
   */
  public abstract String        getMapperDesc();
  
  /**
   * Called to tell the mapper that power is now on.
   */
  public abstract void          powerOn();
  
  /**
   * Called to tell the mapper that power is now off.
   */
  public abstract void          powerOff();
  
  /**
   * Called to tell the mapper to do a cycle.
   */
  public abstract void          doCycle();
}